
from collections import defaultdict


from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
from torch.nn import Parameter
from typing import List, Tuple, Any, Union
import numpy as np

__all__ = ["GNormBatchNorm"]


class GNormBatchNorm(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,
                 ):
        r"""

        Batch normalization for generic representations.

        .. todo ::
            Add more details about how stats are computed and how affine transformation is done.

        Args:
            in_type (FieldType): the input field type
            eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
            momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                    Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
            affine (bool, optional): if ``True``, this module has learnable affine parameters. Default: ``True``

        """
    
        assert isinstance(in_type.gspace, GSpace)
        
        super(GNormBatchNorm, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.affine = affine
        
        self._nfields = None
        
        # group fields by their type and
        #   - check if fields of the same type are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each type
        self._nfields = defaultdict(int)
        
        # indices of the channels corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        ntrivials = 0
        position = 0
        last_field = None
        for i, r in enumerate(self.in_type.representations):
            
            for irr in r.irreps:
                if self.in_type.fibergroup._irreps[irr].is_trivial():
                    ntrivials += 1
            
            if r.name != last_field:
                if not r.name in self._contiguous:
                    self._contiguous[r.name] = True
                else:
                    self._contiguous[r.name] = False

            last_field = r.name
            _indices[r.name] += list(range(position, position + r.size))
            self._nfields[r.name] += 1
            position += r.size
        
        for name, contiguous in self._contiguous.items():
            
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[name] = [min(_indices[name]), max(_indices[name])+1]
                setattr(self, f"{self._escape_name(name)}_indices", _indices[name])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[name] = torch.LongTensor(_indices[name])
                
                # register the indices tensors as parameters of this module
                self.register_buffer(f"{self._escape_name(name)}_indices", _indices[name])
        
        # store the size of each field type
        self._sizes = []
        
        # store for each field type the indices of the trivial irreps in it
        self._trivial_idxs = {}

        # store for each field type the sizes and the indices of all its irreps, grouped by their size
        self._irreps_sizes = {}
        
        for r in self.in_type._unique_representations:
            p = 0
            irreps = defaultdict(lambda: [])
            trivials = []
            aggregator = torch.zeros(r.size, len(r.irreps))
            
            for i, irr in enumerate(r.irreps):
                irr = self.in_type.fibergroup._irreps[irr]
                if irr.is_trivial():
                    trivials.append(p)
                
                aggregator[p:p+irr.size, i] = 1. / irr.size
                
                irreps[irr.size] += list(range(p, p+irr.size))
                p += irr.size
            
            propagator = (aggregator > 0).clone().to(dtype=torch.float)
            
            name = r.name
            
            self._trivial_idxs[name] = torch.tensor(trivials, dtype=torch.long)
            self._irreps_sizes[name] = [(s, idxs) for s, idxs in irreps.items()]
            self._sizes.append((name, r.size))
            
            if not np.allclose(r.change_of_basis, np.eye(r.size)):
                self.register_buffer(f'{self._escape_name(name)}_change_of_basis', torch.tensor(r.change_of_basis, dtype=torch.float))
                self.register_buffer(f'{self._escape_name(name)}_change_of_basis_inv', torch.tensor(r.change_of_basis_inv, dtype=torch.float))
            
            self.register_buffer(f'vars_aggregator_{self._escape_name(name)}', aggregator)
            self.register_buffer(f'vars_propagator_{self._escape_name(name)}', propagator)
        
            running_var = torch.ones((self._nfields[r.name], len(r.irreps)), dtype=torch.float)
            running_mean = torch.zeros((self._nfields[r.name], len(trivials)), dtype=torch.float)
            self.register_buffer(f'{self._escape_name(name)}_running_var', running_var)
            self.register_buffer(f'{self._escape_name(name)}_running_mean', running_mean)
            
            if self.affine:
                weight = Parameter(torch.ones((self._nfields[r.name], len(r.irreps))), requires_grad=True)
                bias = Parameter(torch.zeros((self._nfields[r.name], len(trivials))), requires_grad=True)
                self.register_parameter(f'{self._escape_name(name)}_weight', weight)
                self.register_parameter(f'{self._escape_name(name)}_bias', bias)
            
        self.register_buffer('num_batches_tracked', torch.tensor(0, dtype=torch.long))
        
        self.eps = eps
        self.momentum = momentum

    def reset_running_stats(self):
        for name, size in self._sizes:
            running_var = getattr(self, f"{self._escape_name(name)}_running_var")
            running_mean = getattr(self, f"{self._escape_name(name)}_running_mean")
            running_var.fill_(1)
            running_mean.fill_(0)
        self.num_batches_tracked.zero_()

    def reset_parameters(self):
        self.reset_running_stats()
        if self.affine:
            for name, size in self._sizes:
                weight = getattr(self, f"{self._escape_name(name)}_weight")
                bias = getattr(self, f"{self._escape_name(name)}_bias")
                weight.data.fill_(1)
                bias.data.fill_(0)
    
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        Apply norm non-linearities to the input feature map
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        exponential_average_factor = 0.0

        if self.training:
            self.num_batches_tracked += 1
            if self.momentum is None:  # use cumulative moving average
                exponential_average_factor = 1.0 / self.num_batches_tracked.item()
            else:  # use exponential moving average
                exponential_average_factor = self.momentum

        coords = input.coords
        input = input.tensor
        shape = input.shape[2:]
        b, c = input.shape[:2]
        
        output = torch.empty_like(input)
        
        # iterate through all field types
        for name, size in self._sizes:
            indices = getattr(self, f"{self._escape_name(name)}_indices")
            
            if self._contiguous[name]:
                slice = input[:, indices[0]:indices[1], ...]
            else:
                slice = input[:, indices, ...]

            slice = slice.view(b, -1, size, *shape)
            
            if hasattr(self, f"{self._escape_name(name)}_change_of_basis_inv"):
                cob_inv = getattr(self, f"{self._escape_name(name)}_change_of_basis_inv")
                slice = torch.einsum("ds,bcsxy->bcdxy", (cob_inv, slice))
                
            if self.training:
                
                # compute the mean and variance of the fields
                means, vars = self._compute_statistics(slice, name)
                
                running_var = getattr(self, f"{self._escape_name(name)}_running_var")
                running_mean = getattr(self, f"{self._escape_name(name)}_running_mean")
                
                running_var *= 1 - exponential_average_factor
                running_var += exponential_average_factor * vars
                
                running_mean *= 1 - exponential_average_factor
                running_mean += exponential_average_factor * means
                
                assert torch.allclose(running_mean, getattr(self, f"{self._escape_name(name)}_running_mean"))
                assert torch.allclose(running_var, getattr(self, f"{self._escape_name(name)}_running_var"))
                
            else:
                vars = getattr(self, f"{self._escape_name(name)}_running_var")
                means = getattr(self, f"{self._escape_name(name)}_running_mean")
                
            if self.affine:
                weight = getattr(self, f"{self._escape_name(name)}_weight")
            else:
                weight = 1.
                
            # compute the scalar multipliers needed
            scales = weight / (vars + self.eps).sqrt()

            # compute the point shifts
            # shifts = bias - self._scale(means, scales, name=name)
            centered = self._shift(slice, -1*means, name=name, out=None)
            normalized = self._scale(centered, scales, name=name, out=None)
            
            if self.affine:
                bias = getattr(self, f"{self._escape_name(name)}_bias")
                normalized = self._shift(normalized, bias, name=name, out=None)
            
            if hasattr(self, f"{self._escape_name(name)}_change_of_basis"):
                cob = getattr(self, f"{self._escape_name(name)}_change_of_basis")
                normalized = torch.einsum("ds,bcsxy->bcdxy", (cob, normalized))
                
            if not self._contiguous[name]:
                output[:, indices, ...] = normalized.view(b, -1, *shape)
            else:
                output[:, indices[0]:indices[1], ...] = normalized.view(b, -1, *shape)

            # if self._contiguous[name]:
            #     slice2 = output[:, indices[0]:indices[1], ...]
            # else:
            #     slice2 = output[:, indices, ...]
            # assert torch.allclose(slice2.view(b, -1, size, h, w), slice), name
            
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(NormBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass

    def _compute_statistics(self, t: torch.Tensor, name: str):
        
        trivial_idxs = self._trivial_idxs[name]
        vars_aggregator = getattr(self, f"vars_aggregator_{self._escape_name(name)}")
        
        b, c, s = t.shape[:3]
        shape = t.shape[3:]

        l = trivial_idxs.numel()
        
        # number of samples in the tensor used to estimate the statistics
        M = (np.prod(shape) if shape else 1)
        N = b * M

        # compute the mean of the trivial fields
        trivial_means = t[:, :, trivial_idxs, ...].view(b, c, l, M).mean(dim=(0, 3), keepdim=False).detach()
        
        # compute the mean of squares of all channels
        vars = (t ** 2).view(b, c, s, M).mean(dim=(0, 3), keepdim=False).detach()
        
        # For the non-trivial fields the mean of the fields is 0, so we can compute the variance as the mean of the
        # norms squared.
        # For trivial channels, we need to subtract the squared mean
        vars[:, trivial_idxs] -= trivial_means**2
        
        # aggregate the squared means of the channels which belong to the same irrep
        vars = torch.einsum("io,ci->co", (vars_aggregator, vars))

        # Correct the estimation of the variance with Bessel's correction
        correction = N/(N-1) if N > 1 else 1.
        vars *= correction
        
        return trivial_means, vars

    def _scale(self, t: torch.Tensor, scales: torch.Tensor, name: str, out: torch.Tensor = None):
        
        if out is None:
            out = torch.empty_like(t)
        
        vars_aggregator = getattr(self, f"vars_propagator_{self._escape_name(name)}")
        
        ndims = len(t.shape[3:])
        scale_shape = (1, scales.shape[0], vars_aggregator.shape[0]) + (1,)*ndims
        # scale all fields
        out[...] = t * torch.einsum("oi,ci->co", (vars_aggregator, scales)).reshape(scale_shape)
        
        return out
    
    def _shift(self, t: torch.Tensor, trivial_bias: torch.Tensor, name: str, out: torch.Tensor = None):
    
        if out is None:
            out = t.clone()
        else:
            out[:] = t
            
        trivial_idxs = self._trivial_idxs[name]
        
        bias_shape = (1,) + trivial_bias.shape + (1,)*(len(t.shape) - 3)
        
        # add bias to the trivial fields
        out[:, :, trivial_idxs, ...] += trivial_bias.view(bias_shape)

        return out

    def _escape_name(self, name: str):
        return name.replace('.', '^')



from __future__ import annotations

import numpy as np

import escnn.group
from escnn.group import Group, GroupElement
from escnn.group import IrreducibleRepresentation
from escnn.group import Representation


def build_adjoint_map(G: Group, adj: GroupElement):
    
    _adj = ~adj
    
    def adj_map(e: GroupElement, adj=adj, _adj=_adj, G=G) -> GroupElement:
        assert e.group == G
        return adj @ e @ _adj
    
    return adj_map


def build_trivial_subgroup_maps(G: Group):

    C = escnn.group.cyclic_group(1)

    def parent_map(e: GroupElement, C=C, G=G) -> GroupElement:
        assert e.group == C
        return G.identity

    def child_map(e: GroupElement, C=C, G=G) -> GroupElement:
        assert e.group == G
        if e == G.identity:
            return C.identity
        else:
            return None

    return parent_map, child_map


def build_trivial_irrep():
    
    def trivial_irrep(element: GroupElement) -> np.ndarray:
        return np.eye(1)
    
    return trivial_irrep


def build_trivial_character():
    def trivial_character(element: GroupElement) -> float:
        return 1.
    
    return trivial_character


def build_identity_map():
    def identity(x):
        return x
    return identity





from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch.nn.functional as F
import torch

from typing import List, Tuple, Any, Union

import math

__all__ = [
    "PointwiseAvgPool",
    "PointwiseAvgPoolAntialiased",
    "PointwiseAvgPool2D",
    "PointwiseAvgPoolAntialiased2D",
]


class PointwiseAvgPool2D(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 kernel_size: Union[int, Tuple[int, int]],
                 stride: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = 0,
                 ceil_mode: bool = False
                 ):
        r"""

        Channel-wise average-pooling: each channel is treated independently.
        This module works exactly as :class:`torch.nn.AvgPool2D`, wrapping it in the
        :class:`~escnn.nn.EquivariantModule` interface.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            kernel_size: the size of the window to take a average over
            stride: the stride of the window. Default value is :attr:`kernel_size`
            padding: implicit zero padding to be added on both sides
            ceil_mode: when ``True``, will use ceil instead of floor to compute the output shape

        """

        assert isinstance(in_type.gspace, GSpace)
        assert in_type.gspace.dimensionality == 2

        # for r in in_type.representations:
        #     assert 'pointwise' in r.supported_nonlinearities, \
        #         """Error! Representation "{}" does not support pointwise non-linearities
        #         so it is not possible to pool each channel independently"""
        
        super(PointwiseAvgPool2D, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type

        if isinstance(kernel_size, int):
            self.kernel_size = (kernel_size, kernel_size)
        else:
            self.kernel_size = kernel_size

        if isinstance(stride, int):
            self.stride = (stride, stride)
        elif stride is None:
            self.stride = self.kernel_size
        else:
            self.stride = stride

        if isinstance(padding, int):
            self.padding = (padding, padding)
        else:
            self.padding = padding
            
        self.ceil_mode = ceil_mode
            
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type
        
        # run the common max-pooling
        output = F.avg_pool2d(input.tensor,
                              kernel_size=self.kernel_size,
                              stride=self.stride,
                              padding=self.padding,
                              ceil_mode=self.ceil_mode)
                
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type.size
    
        b, c, hi, wi = input_shape

        # compute the output shape (see 'torch.nn.MaxPool2D')
        ho = (hi + 2 * self.padding[0] - (self.kernel_size[0] - 1) - 1) / self.stride[0] + 1
        wo = (wi + 2 * self.padding[1] - (self.kernel_size[1] - 1) - 1) / self.stride[1] + 1
        
        if self.ceil_mode:
            ho = math.ceil(ho)
            wo = math.ceil(wo)
        else:
            ho = math.floor(ho)
            wo = math.floor(wo)

        return b, self.out_type.size, ho, wo

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        # this kind of pooling is not really equivariant so we can not test equivariance
        pass
    
    
class PointwiseAvgPoolAntialiased2D(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 sigma: float,
                 stride: Union[int, Tuple[int, int]],
                 # kernel_size: Union[int, Tuple[int, int]] = None,
                 padding: Union[int, Tuple[int, int]] = None,
                 #dilation: Union[int, Tuple[int, int]] = 1,
                 ):
        r"""

        Antialiased channel-wise average-pooling: each channel is treated independently.
        It performs strided convolution with a Gaussian blur filter.
        
        The size of the filter is computed as 3 standard deviations of the Gaussian curve.
        By default, padding is added such that input size is preserved if stride is 1.
        
        Based on `Making Convolutional Networks Shift-Invariant Again <https://arxiv.org/abs/1904.11486>`_.
        
        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.
            
        Args:
            in_type (FieldType): the input field type
            sigma (float): standard deviation for the Gaussian blur filter
            stride: the stride of the window.
            padding: additional zero padding to be added on both sides

        """
        
        super(PointwiseAvgPoolAntialiased2D, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        assert sigma > 0.
        
        filter_size = 2*int(round(3*sigma))+1
        
        self.kernel_size = (filter_size, filter_size)
        
        if isinstance(stride, int):
            self.stride = (stride, stride)
        elif stride is None:
            self.stride = self.kernel_size
        else:
            self.stride = stride

        if padding is None:
            padding = int((filter_size-1)//2)
            
        if isinstance(padding, int):
            self.padding = (padding, padding)
        else:
            self.padding = padding

        # Build the Gaussian smoothing filter

        grid_x = torch.arange(filter_size).repeat(filter_size).view(filter_size, filter_size)
        grid_y = grid_x.t()
        grid = torch.stack([grid_x, grid_y], dim=-1)

        mean = (filter_size - 1) / 2.
        variance = sigma ** 2.

        # setting the dtype is needed, otherwise it becomes an integer tensor
        r = -torch.sum((grid - mean) ** 2., dim=-1, dtype=torch.get_default_dtype())

        # Build the gaussian kernel
        _filter = torch.exp(r / (2 * variance))

        # Normalize
        _filter /= torch.sum(_filter)

        # The filter needs to be reshaped to be used in 2d depthwise convolution
        _filter = _filter.view(1, 1, filter_size, filter_size).repeat((in_type.size, 1, 1, 1))

        self.register_buffer('filter', _filter)
        
        ################################################################################################################
    
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""

        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map

        """
        
        assert input.type == self.in_type
        
        output = F.conv2d(input.tensor, self.filter, stride=self.stride, padding=self.padding, groups=input.shape[1])
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords=None)

    def evaluate_output_shape(self, input_shape: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type.size
    
        b, c, hi, wi = input_shape
    
        # compute the output shape
        ho = (hi + 2 * self.padding[0] - (self.kernel_size[0] - 1) - 1) / self.stride[0] + 1
        wo = (wi + 2 * self.padding[1] - (self.kernel_size[1] - 1) - 1) / self.stride[1] + 1
    
        ho = math.floor(ho)
        wo = math.floor(wo)
    
        return b, self.out_type.size, ho, wo

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        # this kind of pooling is not really equivariant so we can't test equivariance
        pass


# for backward compatibility
PointwiseAvgPool = PointwiseAvgPool2D
PointwiseAvgPoolAntialiased = PointwiseAvgPoolAntialiased2D


from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from escnn.gspaces import *

from .equivariant_module import EquivariantModule
from .branching_module import BranchingModule
from .merge_module import MergeModule

from typing import List, Tuple, Union, Any

import torch

import numpy as np

__all__ = ["MultipleModule"]


class MultipleModule(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 labels: List[str],
                 modules: List[Tuple[EquivariantModule, Union[str, List[str]]]],
                 reshuffle: int = 0
                 ):
        r"""
        
        Split the input tensor in multiple branches identified by the input ``labels`` and apply to each of them the
        corresponding module in ``modules``
        
        A label is associated to each field in the input type, while ``modules`` assigns a module to apply to
        each label (or set of labels).
        ``modules`` should be a list of pairs, each containing an :class:`~escnn.nn.EquivariantModule` and a label (or a
        list of labels).
        
        During forward, fields are grouped by the labels and the input tensor is split accordingly.
        Then, each subtensor is passed to the corresponding module in ``modules``.
        
        If ``reshuffle`` is set to a positive integer, a copy of the input tensor is first built sorting the
        fields according to the value set:
        
        - 1: fields are sorted by their labels
        
        - 2: fields are sorted by their labels and, then, by their size
        
        - 3: fields are sorted by their labels, by their size and, then, by their type
        
        In this way, fields that need to be retrieved together are contiguous and it is possible to exploit slicing
        to split the tensor.
        By default, ``reshuffle = 0`` which means that no sorting is performed and, so, if input
        fields are not contiguous this layer will use indexing to retrieve sub-tensors.
        
        This modules wraps a :class:`~escnn.nn.BranchingModule` followed by a :class:`~escnn.nn.MergeModule`.
        
        Args:
            in_type (FieldType): the input field type
            labels (list): the list of labels to group the fields
            modules (list): list of modules to apply to the labeled fields
            reshuffle (int, optional): set how to reshuffle the input fields before splitting the tensor.
                                       By default (``0``) no reshuffling is done
            
        """
        
        assert isinstance(in_type.gspace, GSpace)
        
        super(MultipleModule, self).__init__()
        
        self.gspace = in_type.gspace
        
        self.in_type = in_type
        
        all_labels = set(labels)
        
        modules_labels = []
        
        for _, l in modules:
            if isinstance(l, list):
                modules_labels += l
            else:
                modules_labels.append(l)
        
        modules_labels = set(modules_labels)
        
        assert (modules_labels in all_labels) or (modules_labels == all_labels), "Error! Some labels assigned to the modules don't appear among the channels labels"
        
        # print(labels)
        
        reshuffle_level = int(reshuffle)
        self.branching = BranchingModule(in_type, labels, reshuffle=reshuffle_level)

        for module, l in modules:
            if isinstance(l, str):
                assert module.in_type == self.branching.out_type[l], f"Label {l}, branch class and module ({module}) class don't match:\n [{module.in_type}] \n [{self.branching.out_type[l]}]\n"
            else:
                for i, lb in enumerate(l):
                    assert module.in_type[i] == self.branching.out_type[lb], f"Label {lb}, branch class and module ({module}) class [{i}] don't match: \n [{module.in_type[i]}] \n [{self.branching.out_type[lb]}]\n"
        
        self.merging = MergeModule(modules)
        
        self.out_type = self.merging.out_type
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Split the input tensor according to the labels, apply each module to the corresponding input sub-tensors and
        stack the results.
        
        Args:
            input (GeometricTensor): the input GeometricTensor

        Returns:
            the concatenation of the output of each module
            
        """
        
        assert input.type == self.in_type
        
        sub_tensors = self.branching(input)
        
        return self.merging(sub_tensors)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
        
        branches_shapes = self.branching.evaluate_output_shape(input_shape)
        
        out_shape = self.merging.evaluate_output_shape(branches_shapes)
        
        return out_shape

    def check_equivariance(self, atol: float = 2e-6, rtol: float = 1e-5, full_space_action: bool = True) -> List[Tuple[Any, float]]:
        
        if full_space_action:
            
            return super(MultipleModule, self).check_equivariance(atol=atol, rtol=rtol)
        
        else:
            c = self.in_type.size
        
            x = torch.randn(10, c, 9, 9)
            print(c, self.out_type.size)
            print([r.name for r in self.in_type.representations])
            print([r.name for r in self.out_type.representations])
            x = GeometricTensor(x, self.in_type)
        
            errors = []
        
            for el in self.gspace.testing_elements:
                out1 = self(x).transform_fibers(el)
                out2 = self(x.transform_fibers(el))
            
                errs = (out1.tensor - out2.tensor).detach().numpy()
                errs = np.abs(errs).reshape(-1)
                print(el, errs.max(), errs.mean(), errs.var())
                
                if not torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol):
                    tmp = np.abs((out1.tensor - out2.tensor).detach().numpy())
                    tmp = tmp.reshape(out1.tensor.shape[0], out1.tensor.shape[1], -1).max(axis=2)#.mean(axis=0)
                    
                    np.set_printoptions(precision=2, threshold=200000000, suppress=True, linewidth=500)
                    print(tmp.shape)
                    print(tmp)
            
                assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                    'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                        .format(el, errs.max(), errs.mean(), errs.var())
            
                errors.append((el, errs.mean()))
        
            return errors


from escnn.kernels.basis import KernelBasis, AdjointBasis
from escnn.kernels.steerable_basis import SteerableKernelBasis
from escnn.kernels.wignereckart_solver import WignerEckartBasis, RestrictedWignerEckartBasis

from escnn.kernels.polar_basis import GaussianRadialProfile
from escnn.kernels.polar_basis import CircularShellsBasis

from escnn.group import *

import numpy as np

from typing import List, Union, Callable, Dict, Tuple

__all__ = [
    "kernels_SO2_subgroup_act_R2",
    "kernels_O2_subgroup_act_R2",
    "kernels_SO2_act_R2",
    "kernels_O2_act_R2",
    "kernels_CN_act_R2",
    "kernels_DN_act_R2",
    "kernels_Flip_act_R2",
    "kernels_Trivial_act_R2",
]


def kernels_SO2_act_R2(in_repr: Representation, out_repr: Representation,
                       radii: List[float],
                       sigma: Union[List[float], float],
                       maximum_frequency: int = None,
                       filter: Callable[[Dict], bool] = None
                       ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to continuous rotations, modeled by the
    group :math:`SO(2)`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.SO2`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile

    """
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    
    assert isinstance(group, SO2)
    
    radial_profile = GaussianRadialProfile(radii, sigma)

    if maximum_frequency is None:
        max_in_freq = max(freq for freq, in in_repr.irreps)
        max_out_freq = max(freq for freq, in out_repr.irreps)
        maximum_frequency = max_in_freq + max_out_freq

    return SteerableKernelBasis(
        CircularShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id=(None, -1)
    )



def kernels_O2_act_R2(in_repr: Representation, out_repr: Representation,
                      radii: List[float],
                      sigma: Union[List[float], float],
                      maximum_frequency: int = None,
                      axis: float = np.pi / 2,
                      adjoint: np.ndarray = None,
                      filter: Callable[[Dict], bool] = None
                      ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to reflections and continuous rotations, modeled by the
    group :math:`O(2)`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.O2`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    Because :math:`O(2)` contains all rotations, the reflection element of the group can be associated to any reflection
    axis. Reflections along other axes can be obtained by composition with rotations.
    However, a choice of this axis is required to align the basis with respect to the action of the group.

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        axis (float, optional): angle of the axis of the reflection element
        adjoint (~numpy.ndarray, optional): instead of specifying a reflection axis, you can pass a 2x2 orthogonal
            matrix defining a change of basis on the base space

    """
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    assert isinstance(group, O2)

    radial_profile = GaussianRadialProfile(radii, sigma)

    if maximum_frequency is None:
        max_in_freq = max(freq for _, freq in in_repr.irreps)
        max_out_freq = max(freq for _, freq in out_repr.irreps)
        maximum_frequency = max_in_freq + max_out_freq

    basis = SteerableKernelBasis(
        CircularShellsBasis(maximum_frequency, radial_profile, filter=filter, axis=axis),
        in_repr, out_repr,
        WignerEckartBasis,
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (2, 2)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


###### Automatic subgroups kernel bases

def kernels_O2_subgroup_act_R2(in_repr: Representation, out_repr: Representation,
                               sg_id,
                               radii: List[float],
                               sigma: Union[List[float], float],
                               maximum_frequency: int = 5,
                               axis: float = np.pi / 2.,
                               adjoint: np.ndarray = None,
                               filter: Callable[[Dict], bool] = None
                               ) -> KernelBasis:
    o2 = o2_group(maximum_frequency)
    
    group, _, _ = o2.subgroup(sg_id)
    assert in_repr.group == group
    assert out_repr.group == group
    
    radial_profile = GaussianRadialProfile(radii, sigma)

    basis = SteerableKernelBasis(
        CircularShellsBasis(maximum_frequency, radial_profile, filter=filter, axis=axis),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id=sg_id
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (2, 2)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


def kernels_SO2_subgroup_act_R2(in_repr: Representation, out_repr: Representation,
                                sg_id,
                                radii: List[float],
                                sigma: Union[List[float], float],
                                maximum_frequency: int = 5,
                                adjoint: np.ndarray = None,
                                filter: Callable[[Dict], bool] = None
                                ) -> KernelBasis:
    so2 = so2_group(maximum_frequency)
    
    group, _, _ = so2.subgroup(sg_id)
    assert in_repr.group == group
    assert out_repr.group == group

    o2 = o2_group(maximum_frequency)
    sg_id = o2._combine_subgroups((None, -1), sg_id)

    radial_profile = GaussianRadialProfile(radii, sigma)

    basis = SteerableKernelBasis(
        CircularShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id=sg_id
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (2, 2)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


###### Discrete Symmetries


def kernels_CN_act_R2(in_repr: Representation, out_repr: Representation,
                      radii: List[float],
                      sigma: Union[List[float], float],
                      maximum_frequency: int = None,
                      max_offset: int = None,
                      filter: Callable[[Dict], bool] = None
                      ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to :math:`N` discrete rotations, modeled by
    the group :math:`C_N`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.CyclicGroup`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    The analytical angular solutions of kernel constraints belong to an infinite dimensional space and so can be
    expressed in terms of infinitely many basis elements, each associated with one unique frequency. Because the kernels
    are then sampled on a finite number of points (e.g. the cells of a grid), only low-frequency solutions needs to be
    considered. This enables us to build a finite dimensional basis containing only a finite subset of all analytical
    solutions. ``maximum_frequency`` is an integer controlling the highest frequency sampled in the basis.

    Frequencies also appear in a basis with a period of :math:`N`, i.e. if the basis contains an element with frequency
    :math:`k`, then it also contains an element with frequency :math:`k + N`.
    In the analytical solutions shown in Table 11 `here <https://arxiv.org/abs/1911.08251>`_, each solution has a
    parameter :math:`t` or :math:`\hat{t}`.
    ``max_offset`` defines the maximum absolute value of these two numbers.

    Either ``maximum_frequency`` or ``max_offset`` must be specified.


    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        maximum_frequency (int): maximum frequency of the basis
        max_offset (int): maximum offset in the frequencies of the basis

    """
    
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    
    assert isinstance(group, CyclicGroup)
    
    prefilter = filter
    if max_offset is not None and prefilter is not None:
        filter = lambda attr, max_offset=max_offset, prefilter=prefilter: (attr['j'][0] - attr['_j'][0] <= max_offset) and prefilter(attr)
    elif max_offset is not None and prefilter is None:
        filter = lambda attr, max_offset=max_offset: (attr['j'][0] - attr['_j'][0] <= max_offset)
    else:
        filter = prefilter

    sg_id = group.order()
    return kernels_SO2_subgroup_act_R2(
        in_repr, out_repr, sg_id, radii, sigma, maximum_frequency=maximum_frequency, adjoint=None, filter=filter
    )


def kernels_DN_act_R2(in_repr: Representation, out_repr: Representation,
                      radii: List[float],
                      sigma: Union[List[float], float],
                      axis: float = np.pi / 2,
                      maximum_frequency: int = None,
                      max_offset: int = None,
                      adjoint: np.ndarray = None,
                      filter: Callable[[Dict], bool] = None
                      ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to reflections and :math:`N` discrete rotations,
    modeled by the group :math:`D_N`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s
    of :class:`~escnn.group.DihedralGroup`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    The parameter ``axis`` is the angle in radians (with respect to the horizontal axis, rotating counter-clockwise)
    which defines the reflection axis for the reflection element of the group.

    Frequencies also appear in a basis with a period of :math:`N`, i.e. if the basis contains an element with frequency
    :math:`k`, then it also contains an element with frequency :math:`k + N`.
    In the analytical solutions shown in Table 12 `here <https://arxiv.org/abs/1911.08251>`_, each solution has a
    parameter :math:`t` or :math:`\hat{t}`.
    ``max_offset`` defines the maximum absolute value of these two numbers.

    Either ``maximum_frequency`` or ``max_offset`` must be specified.


    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        maximum_frequency (int): maximum frequency of the basis
        max_offset (int): maximum offset in the frequencies of the basis
        axis (float): angle defining the reflection axis
        adjoint (~numpy.ndarray, optional): instead of specifying a reflection axis, you can pass a 2x2 orthogonal
            matrix defining a change of basis on the base space


    """
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    
    assert isinstance(group, DihedralGroup)

    N = group.rotation_order

    prefilter = filter
    if max_offset is not None and prefilter is not None:
        filter = lambda attr, max_offset=max_offset, prefilter=prefilter: (attr['j'][1] - attr['_j'][1] <= max_offset) and prefilter(attr)
    elif max_offset is not None and prefilter is None:
        filter = lambda attr, max_offset=max_offset: (attr['j'][1] - attr['_j'][1] <= max_offset)
    else:
        filter = prefilter

    sg_id = (0., N)
    return kernels_O2_subgroup_act_R2(
        in_repr, out_repr, sg_id, radii, sigma, axis=axis, maximum_frequency=maximum_frequency, adjoint=adjoint, filter=filter
    )


def kernels_Flip_act_R2(in_repr: Representation, out_repr: Representation,
                        radii: List[float],
                        sigma: Union[List[float], float],
                        axis: float = np.pi / 2,
                        maximum_frequency: int = None,
                        adjoint: np.ndarray = None,
                        filter: Callable[[Dict], bool] = None
                        ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to reflections.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.CyclicGroup`
    with ``N=2``.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    The parameter ``axis`` is the angle in radians (with respect to the horizontal axis, rotating counter-clockwise)
    which defines the reflection axis.

    The analytical angular solutions of kernel constraints belong to an infinite dimensional space and so can be
    expressed in terms of infinitely many basis elements. Only a finite subset can however be implemented.
    ``maximum_frequency`` defines the maximum frequency of a finite-dimensional bandlimited subspace and is therefore
    necessary to specify it.
    See :func:`~escnn.kernels.kernels_CN_act_R2` for more details.


    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        axis (float): angle defining the reflection axis
        maximum_frequency (int): maximum frequency of the basis
        adjoint (~numpy.ndarray, optional): instead of specifying a reflection axis, you can pass a 2x2 orthogonal
            matrix defining a change of basis on the base space

    """
    assert in_repr.group == out_repr.group
    group = in_repr.group
    assert isinstance(group, CyclicGroup) and group.order() == 2
    
    sg_id = (0., 1)
    return kernels_O2_subgroup_act_R2(
        in_repr, out_repr, sg_id, radii, sigma, axis=axis, maximum_frequency=maximum_frequency, adjoint=adjoint, filter=filter
    )


def kernels_Trivial_act_R2(in_repr: Representation, out_repr: Representation,
                           radii: List[float],
                           sigma: Union[List[float], float],
                           maximum_frequency: int = None,
                           filter: Callable[[Dict], bool] = None
                           ) -> KernelBasis:
    r"""

    Builds a basis for unconstrained convolutional kernels.

    This is equivalent to use :func:`~escnn.kernels.kernels_CN_act_R2` with an instance of
    :class:`~escnn.group.CyclicGroup` with ``N=1`` (the trivial group :math:`C_1`).

    ``in_repr`` and ``out_repr`` need to be associated with an instance of :class:`~escnn.group.CyclicGroup` with
    ``N=1``.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    The analytical angular solutions of kernel constraints belong to an infinite dimensional space and so can be
    expressed in terms of infinitely many basis elements. Only a finite subset can however be implemented.
    ``maximum_frequency`` defines the maximum frequency of a finite-dimensinal bandlimited subspace and is therefore
    necessary to specify it.
    See :func:`~escnn.kernels.kernels_CN_act_R2` for more details.

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        axis (float): angle defining the reflection axis
        maximum_frequency (int): maximum frequency of the basis

    """
    
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    assert isinstance(group, CyclicGroup) and group.order() == 1

    sg_id = 1
    return kernels_SO2_subgroup_act_R2(
        in_repr, out_repr, sg_id, radii, sigma, maximum_frequency=maximum_frequency, adjoint=None, filter=filter,
    )







from typing import List, Tuple, Any

import numpy as np

from collections import defaultdict

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule
from .gated1 import GATED_ID, GATES_ID

import torch

from torch.nn import Parameter


__all__ = ["InducedGatedNonLinearity1"]


class InducedGatedNonLinearity1(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 gates: List = None,
                 drop_gates: bool = True,
                 **kwargs
                 ):
        r"""
        
        Induced Gated non-linearities.
        
        .. todo::
            complete documentation!
        
        .. note::
            Make sure all induced gate and gates have same subgroup
            
        
        Args:
            in_type (FieldType): the input field type
            gates (list, optional): list of strings specifying which field in input is a gate and which is a gated field
            drop_gates (bool, optional): if ``True`` (default), drop the trivial fields after using them to compute
                    the gates. If ``False``, the gates are stacked with the gated fields in the output
            
        """

        assert isinstance(in_type.gspace, GSpace)
        
        if gates is None:
            assert len(in_type) % 2 == 0
            
            g = len(in_type) // 2
            gates = [GATES_ID]*g + [GATED_ID]*g
        
        assert len(gates) == len(in_type)
        
        super(InducedGatedNonLinearity1, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type

        self.drop_gates = drop_gates
        
        self._contiguous = {}
        _input_indices = defaultdict(lambda: [])
        _output_indices = defaultdict(lambda: [])

        self._nfields = defaultdict(int)
        
        self.branching = None
        
        for g, r in zip(gates, in_type.representations):
            if g == GATES_ID:
                assert "induced_gate" in r.supported_nonlinearities, \
                    "Error! Representation \"{}\" can't be a \"gate\"".format(r.name)
            elif g == GATED_ID:
                for nl in r.supported_nonlinearities:
                    if nl.startswith("induced_gated"):
                        break
                else:
                    raise ValueError('Error! Representation "{}" does not support "gated" non-linearity'.format(r.name))
            else:
                raise ValueError('Error! "{}" type not recognized'.format(g))

        ngates = len([g for g in gates if g == GATES_ID])
        ngated = len([g for g in gates if g == GATED_ID])

        assert ngates == ngated, \
            'Error! Number of gates ({}) does not match the number of gated non-linearities required ({})' \
            .format(ngates, ngated)

        quotient_size = None
        for g, r in zip(gates, in_type):
            if g == GATES_ID:
                if quotient_size is None:
                    quotient_size = r.size
                else:
                    assert r.size == quotient_size

        subfield_sizes = {}
        for g, r in zip(gates, in_type):
            if g == GATED_ID:
                subfield_size = None
                for nl in r.supported_nonlinearities:
                    if nl.startswith('induced_gated'):
                        assert subfield_size is None, "Error! The representation supports multiple " \
                                                      "sub-fields of different sizes"
                        subfield_size = int(nl.split('_')[-1])
                        assert r.size % subfield_size == 0
                        assert r.size // subfield_size == quotient_size
                        subfield_sizes[r.name] = subfield_size
        
        self.quotient_size = quotient_size
        
        if self.drop_gates:
            # only gated fields are preserved
            # therefore, the output representation is computed from the input one, removing the gates
            self.out_type = in_type.index_select([i for i, g in enumerate(gates) if g == GATED_ID])
        else:
            self.out_type = in_type

        in_last_position = 0
        out_last_position = 0
        last_type = None

        # group fields by their type (gated or gate) and their size, check if fields of the same type are
        # contiguous and retrieve the indices of the fields
        for g, r in zip(gates, in_type.representations):
            if g == GATES_ID:
                type = g
            else:
                type = r.size, subfield_sizes[r.name]
                self._nfields[type] += 1

            if type != last_type:
                if not type in self._contiguous:
                    self._contiguous[type] = True
                else:
                    self._contiguous[type] = False
            last_type = type
    
            _input_indices[type] += list(range(in_last_position, in_last_position + r.size))
            in_last_position += r.size
                
            if g != GATES_ID or not self.drop_gates:
                # since gates are discarded in output, the position on the output fiber is shifted
                # only when a gated field is met
                _output_indices[type] += list(range(out_last_position, out_last_position + r.size))
                out_last_position += r.size
    
        _input_indices = dict(_input_indices)
        # if self.drop_gates:
        _output_indices = dict(_output_indices)
        # else:
        #     self._output_indices = self._input_indices

        for t, contiguous in self._contiguous.items():
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _input_indices[t] = torch.LongTensor([min(_input_indices[t]), max(_input_indices[t]) + 1])
                if t != GATES_ID or not self.drop_gates:
                    _output_indices[t] = torch.LongTensor([min(_output_indices[t]), max(_output_indices[t]) + 1])
            else:
                # otherwise, transform the list of indices into a tensor
                _input_indices[t] = torch.LongTensor(_input_indices[t])
    
                if t != GATES_ID or not self.drop_gates:
                    _output_indices[t] = torch.LongTensor(_output_indices[t])
                    
            # register the indices tensors as parameters of this module
            self.register_buffer('input_indices_{}'.format(t), _input_indices[t])
            if t != GATES_ID or not self.drop_gates:
                self.register_buffer('output_indices_{}'.format(t), _output_indices[t])

        # gates need to be distinguished from gated fields
        _gates_indices = _input_indices.pop(GATES_ID)
        self.register_buffer('gates_indices', _gates_indices)
        
        # build a sorted list of the fields groups, such that every time they are iterated through in the same order
        self._order = sorted(_input_indices.keys())
        
        # the bias for the gates
        self.bias = Parameter(torch.randn(1, ngates, 1, dtype=torch.float), requires_grad=True)

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Apply the gated non-linearity to the input feature map.
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert isinstance(input, GeometricTensor)
        assert input.type == self.in_type
        
        # retrieve the gates
        
        if self._contiguous[GATES_ID]:
            gates = input.tensor[:, self.gates_indices[0]:self.gates_indices[1], ...]
        else:
            gates = input.tensor[:, self.gates_indices, ...]

        coords = input.coords
        
        # retrieving only gated fields from the joint tensor is worthless
        input = input.tensor

        b, c = input.shape[:2]
        spatial_dims = input.shape[2:]

        # transform the gates
        gates = torch.sigmoid(
            gates.view(b, -1, self.quotient_size, *spatial_dims) -
            self.bias.view(1, -1, 1, *[1]*len(spatial_dims))
        )

        # build the output tensor
        output = torch.empty(b, self.out_type.size, *spatial_dims, dtype=torch.float, device=self.bias.device)

        if not self.drop_gates:
            # copy the gates in the output
            if self._contiguous[GATES_ID]:
                output[:, self.gates_indices[0]:self.gates_indices[1], ...] = gates
            else:
                output[:, self.gates_indices, ...] = gates
        
        next_gate = 0
        
        # for each field size
        for type in self._order:
            size, subfield_size = type
            
            # retrieve the needed gates
            g = gates[:, next_gate:next_gate + self._nfields[type], ...].view(b, -1, 1, *spatial_dims)
            
            input_indices = getattr(self, f"input_indices_{type}")
            output_indices = getattr(self, f"output_indices_{type}")
            
            if self._contiguous[type]:
                # if the fields were contiguous, we can use slicing
                output[:, output_indices[0]:output_indices[1], ...] =\
                    (
                            input[:, input_indices[0]:input_indices[1], ...]
                            .view(b, -1, subfield_size, *spatial_dims)
                            * g
                    ).view(b, -1, *spatial_dims)
            
            else:
                # otherwise we have to use indexing
                output[:, output_indices, ...] = \
                    (
                            input[:, input_indices, ...]
                            .view(b, -1, subfield_size, *spatial_dims)
                            * g
                    ).view(b, -1, *spatial_dims)
            
            # shift the position on the gates fiber
            next_gate += self._nfields[type]
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size

        x = torch.randn(3, c, 10, 10)

        x = GeometricTensor(x, self.in_type)

        errors = []

        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
    
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
    
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
    
            errors.append((el, errs.mean()))

        return errors





import os
__cache_path__ = os.path.join(os.path.dirname(__file__), '_cache')


from .utils import psi, chi, psichi

from .group import Group, GroupElement

from .representation import Representation
from .irrep import IrreducibleRepresentation
from .representation import directsum
from .representation import disentangle
from .representation import change_basis
from .representation import homomorphism_space

from .homspace import HomSpace

from .directproduct import DirectProductGroup, direct_product
from .doublegroup import DoubleGroup, double_group

from .groups.factory import *
from .groups.cyclicgroup import CyclicGroup
from .groups.dihedralgroup import DihedralGroup
from .groups.so2group import SO2
from .groups.o2group import O2
from .groups.so3group import SO3
from .groups.o3group import O3
from .groups.ico import Icosahedral
from .groups.octa import Octahedral


from . import _numerical
from . import _clebsh_gordan
from ._clebsh_gordan import InsufficientIrrepsException, UnderconstrainedCGSystem



__all__ = [
    "Group",
    "GroupElement",
    #
    'DirectProductGroup',
    'direct_product',
    #
    'DoubleGroup',
    'double_group',
    # Groups
    "CyclicGroup",
    "DihedralGroup",
    "SO2",
    "O2",
    "SO3",
    "O3",
    "Icosahedral",
    "Octahedral",
    ############
    "trivial_group",
    "cyclic_group",
    "dihedral_group",
    "so2_group",
    "o2_group",
    "so3_group",
    "o3_group",
    "ico_group",
    "octa_group",
    "full_ico_group",
    "full_octa_group",
    "full_cylinder_group",
    "cylinder_group",
    "full_cylinder_discrete_group",
    "cylinder_discrete_group",
    # Representations
    "Representation",
    "IrreducibleRepresentation",
    "directsum",
    "disentangle",
    "change_basis",
    "homomorphism_space",
    #########
    "HomSpace",
    # Exceptions raised by numerical methods
    "InsufficientIrrepsException",
    "UnderconstrainedCGSystem",
    # utils
    "psi",
    "chi",
    "psichi",
    #
    "__cache_path__",
]


groups_dict = {
    CyclicGroup.__name__: CyclicGroup,
    DihedralGroup.__name__: DihedralGroup,
    SO2.__name__: SO2,
    O2.__name__: O2,
    SO3.__name__: SO3,
    O3.__name__: O3,
    Icosahedral.__name__: Icosahedral,
    Octahedral.__name__: Octahedral,
    DirectProductGroup.__name__: DirectProductGroup,
    DoubleGroup.__name__: DoubleGroup,
}


from escnn.kernels.basis import KernelBasis, AdjointBasis, UnionBasis, EmptyBasisException
from escnn.kernels.steerable_basis import SteerableKernelBasis
from escnn.kernels.wignereckart_solver import WignerEckartBasis, RestrictedWignerEckartBasis
from escnn.kernels.sparse_basis import SparseOrbitBasis, SparseOrbitBasisWithIcosahedralSymmetry

from escnn.kernels.polar_basis import GaussianRadialProfile
from escnn.kernels.polar_basis import SphericalShellsBasis

from escnn.group import *

import numpy as np

from typing import List, Union, Callable, Dict, Tuple

__all__ = [
    "kernels_SO3_act_R3",
    "kernels_O3_act_R3",
    ###################
    "kernels_SO3_subgroup_act_R3",
    "kernels_O3_subgroup_act_R3",
    ###################
    "kernels_Ico_act_R3",
    "kernels_Octa_act_R3",
    "kernels_Tetra_act_R3",
    "kernels_FullIco_act_R3",
    "kernels_FullOcta_act_R3",
    "kernels_FullTetra_act_R3",
    "kernels_Pyrito_act_R3",
    ###################
    "kernels_SO2_act_R3",
    "kernels_CN_act_R3",
    "kernels_O2_conical_act_R3",
    "kernels_DN_conical_act_R3",
    "kernels_O2_dihedral_act_R3",
    "kernels_DN_dihedral_act_R3",
    "kernels_full_cylinder_act_R3",
    "kernels_full_cylinder_discrete_act_R3",
    "kernels_cylinder_act_R3",
    "kernels_cylinder_discrete_act_R3",
    "kernels_Inv_act_R3",
    "kernels_Trivial_act_R3",
    ##################
    'kernels_aliased_Ico_act_R3_dodecahedron',
    'kernels_aliased_Ico_act_R3_icosidodecahedron',
    'kernels_aliased_Ico_act_R3_icosahedron',
]


def kernels_SO3_act_R3(in_repr: Representation, out_repr: Representation,
                       radii: List[float],
                       sigma: Union[List[float], float],
                       maximum_frequency: int = None,
                       adjoint: np.ndarray = None,
                       filter: Callable[[Dict], bool] = None
                       ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to continuous rotations, modeled by the
    group :math:`SO(3)`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.SO3`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        adjoint (~numpy.ndarray, optional): 3x3 orthogonal matrix defining a change of basis on the base space

    """
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    
    assert isinstance(group, SO3)

    radial_profile = GaussianRadialProfile(radii, sigma)

    if maximum_frequency is None:
        max_in_freq = max(freq for freq, in in_repr.irreps)
        max_out_freq = max(freq for freq, in out_repr.irreps)
        maximum_frequency = max_in_freq + max_out_freq

    basis = SteerableKernelBasis(
        SphericalShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id='so3'
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (3, 3)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


def kernels_O3_act_R3(in_repr: Representation, out_repr: Representation,
                      radii: List[float],
                      sigma: Union[List[float], float],
                      maximum_frequency: int = None,
                      adjoint: np.ndarray = None,
                      filter: Callable[[Dict], bool] = None
                      ) -> KernelBasis:
    r"""

    Builds a basis for convolutional kernels equivariant to reflections and continuous rotations, modeled by the
    group :math:`O(3)`.
    ``in_repr`` and ``out_repr`` need to be :class:`~escnn.group.Representation` s of :class:`~escnn.group.O3`.

    Because the equivariance constraints allow any choice of radial profile, we use a
    :class:`~escnn.kernels.GaussianRadialProfile`.
    ``radii`` specifies the radial distances at which the rings are centered while ``sigma`` contains the width of each
    of these rings (see :class:`~escnn.kernels.GaussianRadialProfile`).

    Because :math:`O(3)` contains all rotations, the reflection element of the group can be associated to any reflection
    axis.

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field
        radii (list): radii of the rings defining the basis for the radial profile
        sigma (list or float): widths of the rings defining the basis for the radial profile
        adjoint (~numpy.ndarray, optional): 3x3 orthogonal matrix defining a change of basis on the base space

    """
    assert in_repr.group == out_repr.group
    
    group = in_repr.group
    assert isinstance(group, O3)

    radial_profile = GaussianRadialProfile(radii, sigma)

    if maximum_frequency is None:
        max_in_freq = max(freq for _, freq in in_repr.irreps)
        max_out_freq = max(freq for _, freq in out_repr.irreps)
        maximum_frequency = max_in_freq + max_out_freq

    basis = SteerableKernelBasis(
        SphericalShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        WignerEckartBasis,
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (3, 3)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


###### Automatic subgroups kernel bases

def kernels_O3_subgroup_act_R3(in_repr: Representation, out_repr: Representation,
                               sg_id: Tuple,
                               radii: List[float],
                               sigma: Union[List[float], float],
                               maximum_frequency: int = 5,
                               adjoint: np.ndarray = None,
                               filter: Callable[[Dict], bool] = None
                               ) -> KernelBasis:
    
    o3 = o3_group(maximum_frequency)
    
    group, _, _ = o3.subgroup(sg_id)
    assert in_repr.group == group
    assert out_repr.group == group
    
    radial_profile = GaussianRadialProfile(radii, sigma)

    basis = SteerableKernelBasis(
        SphericalShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id=sg_id
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (3, 3)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


def kernels_SO3_subgroup_act_R3(in_repr: Representation, out_repr: Representation,
                                sg_id: Tuple,
                                radii: List[float],
                                sigma: Union[List[float], float],
                                maximum_frequency: int = 5,
                                adjoint: np.ndarray = None,
                                filter: Callable[[Dict], bool] = None
                                ) -> KernelBasis:
    
    so3 = so3_group(maximum_frequency)
    
    group, _, _ = so3.subgroup(sg_id)
    assert in_repr.group == group
    assert out_repr.group == group

    o3 = o3_group(maximum_frequency)
    sg_id = o3._combine_subgroups('so3', sg_id)
    
    radial_profile = GaussianRadialProfile(radii, sigma)

    basis = SteerableKernelBasis(
        SphericalShellsBasis(maximum_frequency, radial_profile, filter=filter),
        in_repr, out_repr,
        RestrictedWignerEckartBasis,
        sg_id=sg_id
    )

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (3, 3)
        basis = AdjointBasis(basis, adjoint)
    
    return basis


###### Platonic Symmetries

def kernels_Ico_act_R3(in_repr: Representation, out_repr: Representation,
                       radii: List[float],
                       sigma: Union[List[float], float],
                       maximum_frequency: int = 5,
                       adjoint: np.ndarray = None,
                       filter: Callable[[Dict], bool] = None
                       ) -> KernelBasis:
    sg_id = 'ico'
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_Octa_act_R3(in_repr: Representation, out_repr: Representation,
                        radii: List[float],
                        sigma: Union[List[float], float],
                        maximum_frequency: int = 5,
                        adjoint: np.ndarray = None,
                        filter: Callable[[Dict], bool] = None
                        ) -> KernelBasis:
    sg_id = 'octa'
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_Tetra_act_R3(in_repr: Representation, out_repr: Representation,
                         radii: List[float],
                         sigma: Union[List[float], float],
                         maximum_frequency: int = 5,
                         adjoint: np.ndarray = None,
                         filter: Callable[[Dict], bool] = None
                         ) -> KernelBasis:
    sg_id = 'tetra'
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


###### Icosahedral Symmetry with Aliased samples

def _kernels_aliased_Ico_act_R3(in_repr: Representation, out_repr: Representation,
                                radii: List[float],
                                sigma: Union[List[float], float],
                                sgid: Tuple,
                                adjoint: np.ndarray = None,
                                ) -> KernelBasis:

    group = ico_group()

    assert in_repr.group == group
    assert out_repr.group == group

    if isinstance(sigma, float):
        sigma = [sigma]*len(radii)
    assert len(sigma) == len(radii)

    basis_list = []
    for r, s in zip(radii, sigma):
        attributes = {'radius': r}
        try:
            if np.isclose(r, 0., rtol=1e-7, atol=1e-7):
                basis = SparseOrbitBasis(
                    X=group.homspace(group.subgroup_self_id),
                    action=group.standard_representation,
                    root=np.zeros(3),
                    sigma=s, attributes=attributes
                )
            else:
                change_of_basis = np.eye(3) * r
                basis = SparseOrbitBasisWithIcosahedralSymmetry(
                    X=group.homspace(sgid),
                    sigma=s, attributes=attributes,
                    change_of_basis = change_of_basis
                )

            basis_list.append(
                SteerableKernelBasis(basis, in_repr, out_repr, WignerEckartBasis)
            )
        except EmptyBasisException:
            pass

    basis = UnionBasis(basis_list)

    if adjoint is not None and not np.allclose(adjoint, np.eye(2)):
        assert adjoint.shape == (3, 3)
        basis = AdjointBasis(basis, adjoint)

    return basis


def kernels_aliased_Ico_act_R3_dodecahedron(in_repr: Representation, out_repr: Representation,
                                            radii: List[float],
                                            sigma: Union[List[float], float],
                                            adjoint: np.ndarray = None,
                                            ) -> KernelBasis:

    sgid = (False, 5)
    return _kernels_aliased_Ico_act_R3(
        in_repr, out_repr, radii, sigma, sgid, adjoint
    )


def kernels_aliased_Ico_act_R3_icosahedron(in_repr: Representation, out_repr: Representation,
                                           radii: List[float],
                                           sigma: Union[List[float], float],
                                           adjoint: np.ndarray = None,
                                           ) -> KernelBasis:
    sgid = (False, 3)
    return _kernels_aliased_Ico_act_R3(
        in_repr, out_repr, radii, sigma, sgid, adjoint
    )


def kernels_aliased_Ico_act_R3_icosidodecahedron(in_repr: Representation, out_repr: Representation,
                                                 radii: List[float],
                                                 sigma: Union[List[float], float],
                                                 adjoint: np.ndarray = None,
                                                 ) -> KernelBasis:

    sgid = (False, 2)
    return _kernels_aliased_Ico_act_R3(
        in_repr, out_repr, radii, sigma, sgid, adjoint
    )


###### Full Platonic Symmetries


def kernels_FullIco_act_R3(in_repr: Representation, out_repr: Representation,
                           radii: List[float],
                           sigma: Union[List[float], float],
                           maximum_frequency: int = 5,
                           adjoint: np.ndarray = None,
                           filter: Callable[[Dict], bool] = None
                           ) -> KernelBasis:
    # TODO
    raise NotImplementedError
    # I_h = I x C_2
    
    sg_id = (True, 'ico')
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_FullOcta_act_R3(in_repr: Representation, out_repr: Representation,
                            radii: List[float],
                            sigma: Union[List[float], float],
                            maximum_frequency: int = 5,
                            adjoint: np.ndarray = None,
                            filter: Callable[[Dict], bool] = None
                            ) -> KernelBasis:
    # TODO
    raise NotImplementedError
    # O_h = O x C_2
    
    sg_id = (True, 'octa')
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_FullTetra_act_R3(in_repr: Representation, out_repr: Representation,
                             radii: List[float],
                             sigma: Union[List[float], float],
                             maximum_frequency: int = 5,
                             adjoint: np.ndarray = None,
                             filter: Callable[[Dict], bool] = None
                             ) -> KernelBasis:
    # TODO
    raise NotImplementedError
    # group T_d
    # n.b. this is different from T x C_2
    
    sg_id = 'fulltetra'
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_Pyrito_act_R3(in_repr: Representation, out_repr: Representation,
                          radii: List[float],
                          sigma: Union[List[float], float],
                          maximum_frequency: int = 5,
                          adjoint: np.ndarray = None,
                          filter: Callable[[Dict], bool] = None
                          ) -> KernelBasis:
    # TODO
    raise NotImplementedError
    # T_h = T x C_2
    # n.b. not a symmetry of the tetrahedron
    
    sg_id = (True, 'tetra')
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


###### Planar symmetries


def kernels_SO2_act_R3(in_repr: Representation, out_repr: Representation,
                       radii: List[float],
                       sigma: Union[List[float], float],
                       maximum_frequency: int = 5,
                       adjoint: np.ndarray = None,
                       filter: Callable[[Dict], bool] = None
                       ) -> KernelBasis:

    sg_id = (False, -1)
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_CN_act_R3(in_repr: Representation, out_repr: Representation,
                      radii: List[float],
                      sigma: Union[List[float], float],
                      maximum_frequency: int = 5,
                      max_offset: int = None,
                      adjoint: np.ndarray = None,
                      filter: Callable[[Dict], bool] = None
                      ) -> KernelBasis:
    assert in_repr.group == out_repr.group
    group = in_repr.group
    assert isinstance(group, CyclicGroup)

    # TODO implement max_offset as filter to apply on top later
    
    sg_id = (False, group.order())
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_O2_conical_act_R3(in_repr: Representation, out_repr: Representation,
                              radii: List[float],
                              sigma: Union[List[float], float],
                              maximum_frequency: int = 5,
                              adjoint: np.ndarray = None,
                              filter: Callable[[Dict], bool] = None
                              ) -> KernelBasis:
    sg_id = ('cone', -1)
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_DN_conical_act_R3(in_repr: Representation, out_repr: Representation,
                              radii: List[float],
                              sigma: Union[List[float], float],
                              # heights: List[float],
                              maximum_frequency: int = 5,
                              max_offset: int = None,
                              adjoint: np.ndarray = None,
                              filter: Callable[[Dict], bool] = None
                              ) -> KernelBasis:
    assert in_repr.group == out_repr.group
    group = in_repr.group
    assert isinstance(group, DihedralGroup)
    
    # TODO implement max_offset as filter to apply on top later

    sg_id = ('cone', group.rotation_order)
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_O2_dihedral_act_R3(in_repr: Representation, out_repr: Representation,
                               radii: List[float],
                               sigma: Union[List[float], float],
                               # heights: List[float],
                               maximum_frequency: int = 5,
                               adjoint: np.ndarray = None,
                               filter: Callable[[Dict], bool] = None
                               ) -> KernelBasis:

    sg_id = (True, -1)
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_DN_dihedral_act_R3(in_repr: Representation, out_repr: Representation,
                               radii: List[float],
                               sigma: Union[List[float], float],
                               # heights: List[float],
                               maximum_frequency: int = 5,
                               max_offset: int = None,
                               adjoint: np.ndarray = None,
                               filter: Callable[[Dict], bool] = None
                               ) -> KernelBasis:
    assert in_repr.group == out_repr.group
    
    # assert all(h >= 0. for h in heights)
    # TODO implement max_offset as filter to apply on top later

    group = in_repr.group
    
    assert isinstance(group, DihedralGroup)

    sg_id = (True, group.rotation_order)
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_full_cylinder_act_R3(in_repr: Representation, out_repr: Representation,
                                 radii: List[float],
                                 sigma: Union[List[float], float],
                                 maximum_frequency: int = 5,
                                 adjoint: np.ndarray = None,
                                 filter: Callable[[Dict], bool] = None
                                ) -> KernelBasis:
    sg_id = True, True, -1
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_full_cylinder_discrete_act_R3(in_repr: Representation, out_repr: Representation,
                                          radii: List[float],
                                          sigma: Union[List[float], float],
                                          # heights: List[float],
                                          maximum_frequency: int = 5,
                                          max_offset: int = None,
                                          adjoint: np.ndarray = None,
                                          filter: Callable[[Dict], bool] = None
                                          ) -> KernelBasis:
    group = in_repr.group
    
    # TODO implement max_offset as filter to apply on top later
    
    sg_id = True, True, group.order()//4
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_cylinder_act_R3(in_repr: Representation, out_repr: Representation,
                            radii: List[float],
                            sigma: Union[List[float], float],
                            maximum_frequency: int = 5,
                            adjoint: np.ndarray = None,
                            filter: Callable[[Dict], bool] = None
                           ) -> KernelBasis:
    sg_id = True, False, -1
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_cylinder_discrete_act_R3(in_repr: Representation, out_repr: Representation,
                                     radii: List[float],
                                     sigma: Union[List[float], float],
                                     # heights: List[float],
                                     maximum_frequency: int = 5,
                                     max_offset: int = None,
                                     adjoint: np.ndarray = None,
                                     filter: Callable[[Dict], bool] = None
                                    ) -> KernelBasis:
    group = in_repr.group
    
    # TODO implement max_offset as filter to apply on top later
    
    sg_id = True, False, group.order() // 2
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency,
        adjoint=adjoint, filter=filter
    )


def kernels_Inv_act_R3(in_repr: Representation, out_repr: Representation,
                       radii: List[float],
                       sigma: Union[List[float], float],
                       maximum_frequency: int = 5,
                       filter: Callable[[Dict], bool] = None
                       ) -> KernelBasis:
    sg_id = (True, False, 1)
    return kernels_O3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency, filter=filter
    )


def kernels_Trivial_act_R3(in_repr: Representation, out_repr: Representation,
                           radii: List[float],
                           sigma: Union[List[float], float],
                           maximum_frequency: int = 5,
                           filter: Callable[[Dict], bool] = None
                           ) -> KernelBasis:
    sg_id = so3_group().subgroup_trivial_id
    return kernels_SO3_subgroup_act_R3(
        in_repr, out_repr, sg_id,
        radii=radii, sigma=sigma,
        maximum_frequency=maximum_frequency, filter=filter
    )





from .gpool import GroupPooling, MaxPoolChannels

from .norm import NormPool
from .induced_norm import InducedNormPool

__all__ = [
    "GroupPooling",
    "NormPool",
    "InducedNormPool",
    "MaxPoolChannels"
]




from escnn.kernels import SteerableKernelBasis, EmptyBasisException
from .basismanager import BasisManager

from typing import Callable, Dict, Iterable, Union, Tuple

import torch
import numpy as np

__all__ = ["SingleBlockBasisSampler", "block_basissampler"]


class SingleBlockBasisSampler(torch.nn.Module, BasisManager):
    
    def __init__(self,
                 basis: SteerableKernelBasis,
                 mask: np.ndarray = None
                 ):
        r"""
        
        Basis expansion method for a single contiguous block, i.e. for kernels whose input type and output type contain
        only fields of one type.
        
        Args:
            basis (SteerableKernelBasis): analytical basis to sample
            mask (np.ndarray, optional): binary mask to select only a subset of the basis elements.
                                         By default (``None``), all elements are kept.
            
        """

        super(SingleBlockBasisSampler, self).__init__()
        
        self.basis = basis
        
        if mask is None:
            mask = np.ones(len(basis), dtype=bool)
            
        assert mask.shape == (len(basis),) and mask.dtype == bool
        
        if not mask.any():
            raise EmptyBasisException

        self._mask = mask
        self.basis = basis

        # we need to know the real output size of the basis elements (i.e. without the change of basis and the padding)
        # to perform the normalization
        sizes = []
        for attr in basis:
            sizes.append(attr["shape"][0])

        sizes = torch.tensor(sizes, dtype=torch.float32).reshape(1, 1, 1, -1)
        sizes = sizes[..., self._mask]

        # to normalize the basis
        self.register_buffer(
            'sizes',
            sizes
        )

    def forward(self, points: torch.Tensor) -> torch.Tensor:

        assert len(points.shape) == 2
        basis = self.basis.sample(points)

        # basis has shape (p, k, o, i)
        # permute to (p, o, i, k)
        basis = basis.permute((0, 2, 3, 1))

        if self._mask is not None:
            basis = basis[:, :, :, self._mask]

        # return basis.to(device=device, dtype=points.dtype) * self.sizes
        return basis * self.sizes

    def get_element_info(self, id: int) -> Dict:
        idx = 0
        for i, attr in enumerate(self.basis):
            if self._mask[i]:
                if idx == id:
                    attr["id"] = idx
                    return attr
                else:
                    idx += 1

    def get_basis_info(self) -> Iterable[Dict]:
        idx = 0
        for i, attr in enumerate(self.basis):
            if self._mask[i]:
                attr["id"] = idx
                idx += 1
                yield attr

    def dimension(self) -> int:
        return self._mask.astype(int).sum()
    
    def __eq__(self, other):
        if isinstance(other, SingleBlockBasisSampler
                      ):
            return (
                    self.basis == other.basis and
                    np.all(self._mask == other._mask)
            )
        else:
            return False
    
    def __hash__(self):
        return 10000 * hash(self.basis) + hash(self._mask.tobytes())

# dictionary storing references to already built basis samplers
# when a new filter tensor is built, it is also stored here
# when the same basis is built again (eg. in another layer), the already existing filter tensor is retrieved
_stored_filters = {}


def block_basissampler(basis: SteerableKernelBasis,
                         basis_filter: Callable[[dict], bool] = None,
                         recompute: bool = False
                         ) -> SingleBlockBasisSampler:
    r"""


    Args:
        basis (SteerableKernelBasis): basis defining the space of kernels
        basis_filter (callable, optional): filter for the basis elements. Should take a dictionary containing an
                                           element's attributes and return whether to keep it or not.
        recompute (bool, optional): whether to recompute new bases or reuse, if possible, already built tensors.

    """

    if basis_filter is not None:
        mask = np.zeros(len(basis), dtype=bool)
        for b, attr in enumerate(basis):
            mask[b] = basis_filter(attr)
    else:
        mask = np.ones(len(basis), dtype=bool)

    if not recompute:
        # compute the mask of the sampled basis containing only the elements allowed by the filter
        key = (basis, mask.tobytes())
        if key not in _stored_filters:
            _stored_filters[key] = SingleBlockBasisSampler(basis, mask)

        return _stored_filters[key]
    
    else:
        return SingleBlockBasisSampler(basis, mask)





from __future__ import annotations

from escnn import gspaces
from escnn import kernels

from .utils import rotate_array_2d

from escnn.group import *

import numpy as np

from typing import Tuple, Union, Callable, List

__all__ = [
    "GSpace2D",
    #################
    "rot2dOnR2",
    "flipRot2dOnR2",
    "flip2dOnR2",
    "trivialOnR2",
]


class GSpace2D(gspaces.GSpace):
    
    def __init__(self, sg_id: Tuple, maximum_frequency: int = 6):
        r"""

        A ``GSpace`` tha describes the set (or subset) of reflectional and rotational symmetries of the 2D
        Euclidean Space :math:`\R^2`.
        The subset of symmetries is determined by the subgroup of :math:`\O2` that is specified by `sg_id`
        (check the documentation of :class:`escnn.group.O2`).

        Args:
            sg_id (tuple): The ID of the subgroup within the fiber group :math:`\O2` that determines the reflectional and rotational symmetries to consider. For detailed documentation on the ID of each subgroup, refer to the documentation of :class:`escnn.group.O2`

            maximum_frequency (int): Maximum frequency of the irreps to pre-instantiate, if the symmetry group (identified by `sg_id`) contains all continuous rotations.

        .. note ::
            A point :math:`\bold{v} \in \R^2` is parametrized using an :math:`(X, Y)` convention,
            i.e. :math:`\bold{v} = (x, y)^T`.
            The representation :attr:`escnn.gspaces.GSpace2D.basespace_action` also assumes this convention.
            
            However, when working with data on a pixel grid, the usual :math:`(-Y, X)` convention is used.
            That means that, in a 4-dimensional feature tensor of shape ``(B, C, D1, D2)``, the last dimension
            is the X axis while the second last is the (inverted) Y axis.
            Note that this is consistent with 2D images, where a :math:`(-Y, X)` convention is used.
            
        """
        
        o2 = o2_group(maximum_frequency=maximum_frequency)
        _sg_id = o2._process_subgroup_id(sg_id)
        fibergroup, inclusion, restriction = o2.subgroup(_sg_id)
        
        # TODO - catch sg_id and build a dictionary of more meaningful names
        # use the input sg_id instead of the processed one to avoid adding the adjoint parameter unless specified
        name = f'{fibergroup}_on_R2[{sg_id}]'
        
        self._sg_id = _sg_id
        self._inclusion = inclusion
        self._restriction = restriction
        self._base_action = o2.irrep(1, 1).restrict(_sg_id)

        super(GSpace2D, self).__init__(fibergroup, 2, name)
    
    def restrict(self, id: Tuple) -> Tuple[gspaces.GSpace, Callable, Callable]:
        r"""

        Build the :class:`~escnn.group.GSpace` associated with the subgroup of the current fiber group identified by
        the input ``id``

        Args:
            id (tuple): the id of the subgroup

        Returns:
            a tuple containing

                - **gspace**: the restricted gspace

                - **back_map**: a function mapping an element of the subgroup to itself in the fiber group of the original space

                - **subgroup_map**: a function mapping an element of the fiber group of the original space to itself in the subgroup (returns ``None`` if the element is not in the subgroup)

        """
        
        o2 = o2_group()
        sg_id = o2._combine_subgroups(self._sg_id, id)
        sg, inclusion, restriction = self.fibergroup.subgroup(id)
        
        return GSpace2D(sg_id), inclusion, restriction
    
    @property
    def rotations_order(self):
        return self._sg_id[1]
    
    @property
    def flips_order(self):
        return 1 if self._sg_id[0] is not None else 0

    def _basis_generator(self,
                         in_repr: Representation,
                         out_repr: Representation,
                         rings: List[float],
                         sigma: List[float],
                         **kwargs,
                         ) -> kernels.KernelBasis:
        r"""
        Method that builds the analytical basis that spans the space of equivariant filters which
        are intertwiners between the representations induced from the representation ``in_repr`` and ``out_repr``.
        
        `kwargs` can be used to specify `maximum_frequency`

        Args:
            in_repr (Representation): the input representation
            out_repr (Representation): the output representation
            rings (list): radii of the rings where to sample the bases
            sigma (list): parameters controlling the width of each ring where the bases are sampled.

        Returns:
            the basis built

        """
        
        # TODO - add max_offset for cyclic and dihedral groups!
        
        if 'maximum_frequency' in kwargs:
            maximum_frequency = kwargs['maximum_frequency']
        else:
            maximum_frequency = None
        
        if self._sg_id[0] is not None and self._sg_id[1] == -1:
            return kernels.kernels_O2_act_R2(in_repr, out_repr, rings, sigma, axis=self._sg_id[0]/2, maximum_frequency=maximum_frequency, filter=None)
        elif self._sg_id == (None, -1):
            return kernels.kernels_SO2_act_R2(in_repr, out_repr, rings, sigma, maximum_frequency=maximum_frequency, filter=None)
        elif self._sg_id[0] is None:
            sg_id = self._sg_id[1]
            return kernels.kernels_SO2_subgroup_act_R2(in_repr, out_repr, sg_id, rings, sigma, adjoint=None, maximum_frequency=maximum_frequency, filter=None)
        else:
            return kernels.kernels_O2_subgroup_act_R2(in_repr, out_repr, self._sg_id, rings, sigma, axis=0., adjoint=None, maximum_frequency=maximum_frequency, filter=None)

    @property
    def basespace_action(self) -> Representation:
        return self._base_action

    def __eq__(self, other):
        if isinstance(other, GSpace2D):
            return self._sg_id == other._sg_id
        else:
            return False
    
    def __hash__(self):
        return 1000 * hash(self.name) + hash(self._sg_id)
    
    
def rot2dOnR2(N: int = -1, maximum_frequency: int = 6) -> GSpace2D:
    r"""
    
    Describes rotation symmetries of the plane :math:`\R^2`.

    If ``N > 1``, the gspace models *discrete* rotations by angles which are multiple of :math:`\frac{2\pi}{N}`
    (:class:`~e2cnn.group.CyclicGroup`).
    Otherwise, if ``N=-1``, the gspace models *continuous* planar rotations (:class:`~e2cnn.group.SO2`).
    In that case the parameter ``maximum_frequency`` is required to specify the maximum frequency of the irreps of
    :class:`~e2cnn.group.SO2` (see its documentation for more details)

    Args:
        N (int): number of discrete rotations (integer greater than 1) or ``-1`` for continuous rotations
        maximum_frequency (int): maximum frequency of :class:`~e2cnn.group.SO2`'s irreps if ``N = -1``

    """
    assert isinstance(N, int)
    assert N == -1 or N > 0
    sg_id = None, N
    return GSpace2D(sg_id, maximum_frequency=maximum_frequency)
    
    
def flipRot2dOnR2(N: int = -1, maximum_frequency: int = 6, axis: float = np.pi / 2.) -> GSpace2D:
    r"""
    Describes reflectional and rotational symmetries of the plane :math:`\R^2`.
    
    Reflections are applied with respect to the line through the origin with an angle ``axis`` degrees with respect
    to the *X*-axis.
    
    If ``N > 1``, this gspace models reflections and *discrete* rotations by angles multiple of :math:`\frac{2\pi}{N}`
    (:class:`~e2cnn.group.DihedralGroup`).
    Otherwise, if ``N=-1`` (by default), the class models reflections and *continuous* planar rotations
    (:class:`~e2cnn.group.O2`).
    In that case, the parameter ``maximum_frequency`` is required to specify the maximum frequency of the irreps of
    :class:`~e2cnn.group.O2` (see its documentation for more details)
    
    .. note ::
        
        All axes obtained from the axis defined by ``axis`` with a rotation in the symmetry group are equivalent.
        For instance, if ``N = 4``, an axis :math:`\beta` is equivalent to the axis :math:`\beta + \pi/2`.
        It follows that for ``N = -1``, i.e. in case the symmetry group contains all continuous rotations, any
        reflection axis is theoretically equivalent.
        In practice, though, a basis for equivariant convolutional filter sampled on a grid is affected by the
        specific choice of the axis. In general, choosing an axis aligned with the grid (an horizontal or a
        vertical axis, i.e. :math:`0` or :math:`\pi/2`) is suggested.
    
    Args:
        N (int): number of discrete rotations (integer greater than 1) or -1 for continuous rotations
        maximum_frequency (int): maximum frequency of :class:`~e2cnn.group.O2` 's irreps if ``N = -1``
        axis (float, optional): the slope of the axis of the flip (in radians)
        
    """
    
    assert isinstance(N, int)
    assert N == -1 or N > 0
    sg_id = 2*axis, N
    return GSpace2D(sg_id, maximum_frequency=maximum_frequency)


def flip2dOnR2(axis: float = np.pi / 2) -> GSpace2D:
    r"""
    
    Describes reflectional symmetries of the plane :math:`\R^2`.
    
    Reflections are applied along the line through the origin with an angle ``axis`` degrees with respect to
    the *X*-axis.
    
    Args:
        axis (float, optional): the slope of the axis of the reflection (in radians).
                                By default, the vertical axis is used (:math:`\pi/2`).
                                
    """
    sg_id = 2*axis, 1
    return GSpace2D(sg_id, maximum_frequency=1)


def trivialOnR2() -> GSpace2D:
    r"""
    Describes the plane :math:`\R^2` without considering any origin-preserving symmetry.
    This is modeled by choosing trivial fiber group :math:`\{e\}`.
    
    .. note ::
        This models the symmetries of conventional *Convolutional Neural Networks* which are not equivariant to
        origin preserving transformations such as rotations and reflections.
        
    """
    sg_id = (None, 1)
    return GSpace2D(sg_id, maximum_frequency=1)






from typing import List, Tuple, Any

import numpy as np

from collections import defaultdict

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch

from torch.nn import Parameter


__all__ = ["GatedNonLinearity1", "GATED_ID", "GATES_ID"]


GATED_ID = "gated"
GATES_ID = "gate"


class GatedNonLinearity1(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 gates: List = None,
                 drop_gates: bool = True,
                 **kwargs
                 ):
        r"""
        
        Gated non-linearities.
        This module applies a bias and a sigmoid function of the gates fields and, then, multiplies each gated
        field by one of the gates.
        
        The input representation of the gated fields is preserved by this operation while the gate fields are
        discarded.
        
        The gates and the gated fields are provided in one unique input tensor and, therefore, :attr:`in_repr` should
        be the representation of the fiber containing both gates and gated fields.
        Moreover, the parameter :attr:`gates` needs to be set with a list long as the total number of fields,
        containing in a position ``i`` the string ``"gate"`` if the ``i``-th field is a gate or the string ``"gated"``
        if the ``i``-th field is a gated field. No other strings are allowed.
        By default (``gates = None``), the first half of the fields is assumed to contain the gates (and, so, these
        fields have to be trivial fields) while the second one is assumed to contain the gated fields.
        
        In any case, the number of gates and the number of gated fields have to match (therefore, the number of
        fields has to be an even number).
        
        Args:
            in_type (FieldType): the input field type
            gates (list, optional): list of strings specifying which field in input is a gate and which is a gated field
            drop_gates (bool, optional): if ``True`` (default), drop the trivial fields after using them to compute
                    the gates. If ``False``, the gates are stacked with the gated fields in the output
            
        """

        assert isinstance(in_type.gspace, GSpace)
        
        if gates is None:
            assert len(in_type) % 2 == 0
            
            g = len(in_type) // 2
            gates = [GATES_ID]*g + [GATED_ID]*g
        
        assert len(gates) == len(in_type)
        
        super(GatedNonLinearity1, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type

        self.drop_gates = drop_gates
        
        self._contiguous = {}
        _input_indices = defaultdict(lambda: [])
        _output_indices = defaultdict(lambda: [])

        self._nfields = defaultdict(int)
        
        self.branching = None
        
        for g, r in zip(gates, in_type.representations):
            if g == GATES_ID:
                # assert GATES_ID in r.supported_nonlinearities, \
                assert r.is_trivial(), \
                    "Error! Representation \"{}\" can't be a \"gate\"".format(r.name)
            elif g == GATED_ID:
                assert GATED_ID in r.supported_nonlinearities, \
                    'Error! Representation "{}" does not support "gated" non-linearity'.format(r.name)
            else:
                raise ValueError('Error! "{}" type not recognized'.format(g))

        ngates = len([g for g in gates if g == GATES_ID])
        ngated = len([g for g in gates if g == GATED_ID])

        assert ngates == ngated, \
            'Error! Number of gates ({}) does not match the number of gated non-linearities required ({})' \
            .format(ngates, ngated)

        if self.drop_gates:
            # only gated fields are preserved
            # therefore, the output representation is computed from the input one, removing the gates
            self.out_type = in_type.index_select([i for i, g in enumerate(gates) if g == GATED_ID])
        else:
            self.out_type = in_type

        in_last_position = 0
        out_last_position = 0
        last_type = None

        # group fields by their type (gated or gate) and their size, check if fields of the same type are
        # contiguous and retrieve the indices of the fields
        for g, r in zip(gates, in_type.representations):
            if g == GATES_ID:
                type = g
            else:
                type = r.size
                self._nfields[r.size] += 1

            if type != last_type:
                if not type in self._contiguous:
                    self._contiguous[type] = True
                else:
                    self._contiguous[type] = False
            last_type = type
    
            _input_indices[type] += list(range(in_last_position, in_last_position + r.size))
            in_last_position += r.size
                
            if g != GATES_ID or not self.drop_gates:
                # since gates are discarded in output, the position on the output fiber is shifted
                # only when a gated field is met
                _output_indices[type] += list(range(out_last_position, out_last_position + r.size))
                out_last_position += r.size
    
        _input_indices = dict(_input_indices)
        # if self.drop_gates:
        _output_indices = dict(_output_indices)
        # else:
        #     self._output_indices = self._input_indices

        for t, contiguous in self._contiguous.items():
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _input_indices[t] = torch.LongTensor([min(_input_indices[t]), max(_input_indices[t]) + 1])
                if t != GATES_ID or not self.drop_gates:
                    _output_indices[t] = torch.LongTensor([min(_output_indices[t]), max(_output_indices[t]) + 1])
            else:
                # otherwise, transform the list of indices into a tensor
                _input_indices[t] = torch.LongTensor(_input_indices[t])
    
                if t != GATES_ID or not self.drop_gates:
                    _output_indices[t] = torch.LongTensor(_output_indices[t])
                    
            # register the indices tensors as parameters of this module
            self.register_buffer('input_indices_{}'.format(t), _input_indices[t])
            if t != GATES_ID or not self.drop_gates:
                self.register_buffer('output_indices_{}'.format(t), _output_indices[t])

        # gates need to be distinguished from gated fields
        _gates_indices = _input_indices.pop(GATES_ID)
        self.register_buffer('gates_indices', _gates_indices)
        
        # build a sorted list of the fields groups, such that every time they are iterated through in the same order
        self._order = sorted(_input_indices.keys())
        
        # the bias for the gates
        self.bias = Parameter(torch.randn(1, ngates, dtype=torch.float), requires_grad=True)

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        
        Apply the gated non-linearity to the input feature map.
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert isinstance(input, GeometricTensor)
        assert input.type == self.in_type
        
        # retrieve the gates
        
        if self._contiguous[GATES_ID]:
            gates = input.tensor[:, self.gates_indices[0]:self.gates_indices[1], ...]
        else:
            gates = input.tensor[:, self.gates_indices, ...]

        coords = input.coords
        
        # retrieving only gated fields from the joint tensor is worthless
        input = input.tensor

        b, c = input.shape[:2]
        spatial_shape = input.shape[2:]

        # transform the gates
        gates = torch.sigmoid(gates - self.bias.view(1, -1, *[1]*len(spatial_shape)))

        # build the output tensor
        output = torch.empty(b, self.out_type.size, *spatial_shape, dtype=torch.float, device=self.bias.device)

        if not self.drop_gates:
            # copy the gates in the output
            if self._contiguous[GATES_ID]:
                output[:, self.gates_indices[0]:self.gates_indices[1], ...] = gates
            else:
                output[:, self.gates_indices, ...] = gates
        
        next_gate = 0
        
        # for each field size
        for size in self._order:
            
            # retrieve the needed gates
            g = gates[:, next_gate:next_gate + self._nfields[size], ...].view(b, -1, 1, *spatial_shape)
            
            input_indices = getattr(self, f"input_indices_{size}")
            output_indices = getattr(self, f"output_indices_{size}")
            
            if self._contiguous[size]:
                # if the fields were contiguous, we can use slicing
                output[:, output_indices[0]:output_indices[1], ...] =\
                    (
                            input[:, input_indices[0]:input_indices[1], ...]
                            .view(b, -1, size, *spatial_shape)
                            * g
                    ).view(b, -1, *spatial_shape)
            
            else:
                # otherwise we have to use indexing
                output[:, output_indices, ...] = \
                    (
                            input[:, input_indices, ...]
                            .view(b, -1, size, *spatial_shape)
                            * g
                    ).view(b, -1, *spatial_shape)
            
            # shift the position on the gates fiber
            next_gate += self._nfields[size]
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        
        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size
    
        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)
        
    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size

        x = torch.randn(3, c, *[10]*self.in_type.gspace.dimensionality)

        x = GeometricTensor(x, self.in_type)

        errors = []

        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
    
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
    
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
    
            errors.append((el, errs.mean()))

        return errors





from .basis import EmptyBasisException, KernelBasis, AdjointBasis, UnionBasis

from .harmonic_polynomial_r3 import HarmonicPolynomialR3Generator

from .steerable_filters_basis import SteerableFiltersBasis, PointBasis
from .polar_basis import GaussianRadialProfile, SphericalShellsBasis, CircularShellsBasis
from .sparse_basis import SparseOrbitBasis, SparseOrbitBasisWithIcosahedralSymmetry

from .steerable_basis import SteerableKernelBasis, IrrepBasis

from .wignereckart_solver import WignerEckartBasis, RestrictedWignerEckartBasis

from .r2 import *
from .r3 import *


import escnn.group


def kernels_on_point(in_repr: escnn.group.Representation, out_repr: escnn.group.Representation) -> KernelBasis:
    r"""

    Args:
        in_repr (Representation): the representation specifying the transformation of the input feature field
        out_repr (Representation): the representation specifying the transformation of the output feature field

    """
    assert in_repr.group == out_repr.group

    group = in_repr.group

    basis = SteerableKernelBasis(
        PointBasis(group), in_repr, out_repr,
        WignerEckartBasis,
    )

    return basis


__all__ = [
    "EmptyBasisException",
    "KernelBasis",
    # General Bases
    "AdjointBasis",
    "UnionBasis",
    # Steerable Kernel Bases
    'SteerableKernelBasis',
    "IrrepBasis",
    "WignerEckartBasis",
    "RestrictedWignerEckartBasis",
    # Steerable Filters Bases
    'SteerableFiltersBasis',
    'PointBasis',
    "SphericalShellsBasis",
    "CircularShellsBasis",
    'GaussianRadialProfile',
    "SparseOrbitBasis",
    "SparseOrbitBasisWithIcosahedralSymmetry",
    # Other useful modules
    "HarmonicPolynomialR3Generator",
    # Generic group acting on R^0
    "kernels_on_point",
    # R2 bases
    "kernels_Flip_act_R2",
    "kernels_DN_act_R2",
    "kernels_O2_act_R2",
    "kernels_Trivial_act_R2",
    "kernels_CN_act_R2",
    "kernels_SO2_act_R2",
    "kernels_SO2_subgroup_act_R2",
    "kernels_O2_subgroup_act_R2",
    # R3 bases
    "kernels_O3_act_R3",
    "kernels_SO3_act_R3",
    "kernels_SO3_subgroup_act_R3",
    "kernels_O3_subgroup_act_R3",
    "kernels_Ico_act_R3",
    "kernels_Octa_act_R3",
    "kernels_Tetra_act_R3",
    "kernels_FullIco_act_R3",
    "kernels_FullOcta_act_R3",
    "kernels_FullTetra_act_R3",
    "kernels_Pyrito_act_R3",
    "kernels_SO2_act_R3",
    "kernels_CN_act_R3",
    "kernels_O2_conical_act_R3",
    "kernels_DN_conical_act_R3",
    "kernels_O2_dihedral_act_R3",
    "kernels_DN_dihedral_act_R3",
    "kernels_Inv_act_R3",
    "kernels_Trivial_act_R3",
    'kernels_aliased_Ico_act_R3_dodecahedron',
    'kernels_aliased_Ico_act_R3_icosidodecahedron',
    'kernels_aliased_Ico_act_R3_icosahedron',
]




from __future__ import annotations

from typing import Tuple, Callable, Iterable, List, Any, Dict

import escnn.group

from escnn.group import Group, GroupElement, IrreducibleRepresentation, DirectProductGroup
from escnn.group.irrep import restrict_irrep

import numpy as np
import itertools
import re

__all__ = [
    'DoubleGroup',
    'double_group'
]


class DoubleGroup(DirectProductGroup):

    def __init__(self, G: str, name: str = None, **group_keys):
        r"""

        Class defining the direct product of a group with itself.
        
        This is a special case (a subclass) of :class:`~escnn.group.DirectProductGroup`.
        :class:`~escnn.group.DirectProductGroup` can not automatically support all possible subgroups and, in particular,
        it does not support the diagonal subgroup of :math:`G \times G` which is isomorphic to :math:`G`.
        This subclass supports this special subgroup, which is identified by the id `"diagonal"`.
        See also :meth:`~escnn.group.DoubleGroup.subgroup_diagonal_id`.
        
        .. warning::
            This class should not be directly instantiated to ensure caching is performed correclty.
            You should instead use the function :func:`~escnn.group.double_group`.

        .. warning::
            This class does not support all possible subgroups of the direct product!
            If :math:`G = G_1 \times G_1`, only the diagonal subgroup isomorphic to :math:`G_1` and the subgroups of
            the form :math:`H = H_1 \times H_2` with :math:`H_1 < G_1` and :math:`H_2 < G_2` are supported.

            A subgroup `id` is a tuple containing a pair `(id1, id2)`, where `id1` identifies a subgroup of :math:`G_1`
            while `id2` identifies a subgroup of :math:`G_2`.
            See the documentation of the two groups used for more details about their subgroup structures.

            In case either :math:`H_1` or :math:`H_2` is the trivial subgroup (containing only the identity),
            the subgroup returned is just an instance of the other one and not of
            :class:`~escnn.group.DirectProductGroup`.


        Args:
            G1 (Group): first group
            G2 (Group): second group
            name (str, optional): name assigned to the resulting group
            groups_keys: additional keywords argument used for identifying the groups and perform caching

        Attributes:
            ~.G1 (Group): the first group
            ~.G2 (Group): the second group

        """
    
        assert all(k.startswith('G_') for k in group_keys.keys())
    
        keys1 = {
            'G1_' + k[2:]: v for k, v in group_keys.items()
        }
        keys2 = {
            'G2_' + k[2:]: v for k, v in group_keys.items()
        }
        super(DoubleGroup, self).__init__(G, G, name, **keys1, **keys2)

    @property
    def _keys(self) -> Dict[str, Any]:
        keys = dict()
        keys['G'] = self.G1.__class__.__name__

        if not self._defaulf_name:
            keys['name'] = self.name

        keys.update({
            'G_' + k: v
            for k, v in self.G1._keys.items()
        })
        return keys

    @property
    def subgroup_diagonal_id(self):
        r"""
            The subgroup id associated with the diagonal group.
            This is the subgroup containing elements in the form :math:`(g, g)` for :math:`g \in G` and is isomorphic
            to :math:`G` itself.
            The id can be used in the method :meth:`~escnn.group.Group.subgroup` to generate the subgroup.
        """
        return 'diagonal'

    def _process_subgroup_id(self, id):
        if id == 'diagonal':
            return id
        else:
            return super(DoubleGroup, self)._process_subgroup_id(id)

    def _combine_subgroups(self, sg_id1, sg_id2):
        raise NotImplementedError

    def _subgroup(self, id) -> Tuple[
        escnn.group.Group,
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement],
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement]
    ]:
        if id == 'diagonal':
            return self.G1, inclusion(self), restriction(self)

        else:
            return super(DoubleGroup, self)._subgroup(id)

    _cached_group_instance = dict()

    @classmethod
    def _generator(cls, G: str, **group_keys) -> 'DirectProductGroup':

        key = {
            'G': G,
        }
        key.update(**group_keys)

        key = tuple(sorted(key.items()))

        if key not in cls._cached_group_instance:
            cls._cached_group_instance[key] = DoubleGroup(G, **group_keys)

        cls._cached_group_instance[key]._build_representations()

        return cls._cached_group_instance[key]


def restriction(G: DoubleGroup):
    def _map(e: GroupElement, G=G):
        assert e.group == G

        e1, e2 = G.split_element(e)

        if e1 == e2:
            return e1
        else:
            return None
    return _map


def inclusion(G: DoubleGroup):
    def _map(e: GroupElement, G=G):
        assert e.group == G.G1

        return G.element(
            (e.value, e.value),
            param=f'[{e.param} | {e.param}]'
        )

    return _map


def double_group(G: Group, name: str = None):
    r'''

    Generates the direct product of the input group `G` with itself.

    Args:
        G (Group): the group
        name (str, optional): name assigned to the resulting group

    Returns:
        an instance of :class:`~escnn.group.DoubleGroup`

    '''
    
    group_keys = {
        'G_' + k: v
        for k, v in G._keys.items()
    }
    return DoubleGroup._generator(
        G.__class__.__name__,
        name=name,
        **group_keys
    )





from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Tuple, Callable, Iterable, List, Any, Dict

import escnn.group
import numpy as np
from scipy import sparse


__all__ = ["Group", "GroupElement"]


class Group(ABC):
    
    @property
    @abstractmethod
    def PARAM(self) -> str:
        f"""
            Default parametrization used for storing the elements of the group.
        """
        pass

    @property
    @abstractmethod
    def PARAMETRIZATIONS(self) -> List[str]:
        f"""
            List of all supported parametrizations of the group.
        """
        pass

    def __init__(self, name: str, continuous: bool, abelian: bool):
        r"""
        Abstract class defining the interface of a group.
        
        A group is a set of *group elements* together with a binary operation satisfying a number of axioms.
        
        In this library, this is implemented using this class :class:`~escnn.group.Group` and the class
        :class:`~escnn.group.GroupElement`.
        
        One can retrieve or generate elements of a group by using, for instance, the properties or methods
        :meth:`~escnn.group.Group.identity` , :meth:`~escnn.group.Group.elements` or :meth:`~escnn.group.Group.sample`.
        Each group may also have additional methods to generate its group elements.
        Additionally, one can use the method :meth:`~escnn.group.Group.element` to generate a new group element.
        
        The group algebra is directly implemented inside :class:`~escnn.group.GroupElement` such that one can combine
        group elements in a way that resamples mathematical expressions.
        In particular, the ``@`` implements the binary product while ``~`` implements the group inverse.
        See :class:`~escnn.group.GroupElement` for more details.
        
        
        Args:
            name (str): name identifying the group
            continuous (bool): whether the group is non-finite or finite
            abelian (bool): whether the group is *abelian* (commutative)
            
        Attributes:
            ~.name (str): Name identifying the group
            ~.continuous (bool): Whether it is a non-finite or a finite group
            ~.abelian (bool): Whether it is an *abelian* group (i.e. if the group law is commutative)

        """
        
        self.name = name
        
        self.continuous = continuous
        
        self.abelian = abelian
        
        self._irreps = {}
        
        self._representations = {}
        
        self._subgroups = {}
        
        self._homspaces = {}
        
    def order(self) -> int:
        r"""
        Returns the number of elements in this group if it is a finite group, otherwise -1 is returned
        
        Returns:
            the size of the group or ``-1`` if it is a continuous group

        """
        if self.elements is not None:
            return len(self.elements)
        else:
            return -1

    def element(self, element, param: str = None) -> GroupElement:
        r"""
            Generate the element of the current group parametrized by ``element`` according to
            the parametrization ``param``.
            
            Each group supports a different set of parametrizations.
            By default, the parametrization :attr:`escnn.group.Group.PARAM` is used.
            The list of all available parametrizations of a group can be accessed through the property
            :attr:`escnn.group.Group.PARAMETRIZATIONS`.
            
        Args:
            element: values parametrizing a group element.
            param (str): string identifying the parametrization to be used.

        Returns:
            an instance of :class:`~escnn.group.GroupElement`

        """
        if param is None:
            param = self.PARAM
            
        return GroupElement(element, self, param)
    
    @property
    @abstractmethod
    def identity(self) -> GroupElement:
        r"""
            The identity element of the group.
            
            The identity element :math:`e` satisfies the following property
            :math:`\forall\ g \in G,\ g \cdot e = e \cdot g= g` .
            
        """
        pass

    @property
    @abstractmethod
    def elements(self) -> List[GroupElement]:
        r"""
            If the group is finite (i.e. ``self.continuous = False``), it is a list of all group elements.
            If the group is not finite, this property is `None`.
            
            Returns:
                a list of :class:`~escnn.group.GroupElement` instances
        """
        pass

    @property
    @abstractmethod
    def subgroup_trivial_id(self):
        r"""
            The subgroup `id` associated with the trivial subgroup containing only the identity element :math:`{e}`.
            The id can be used in the method :meth:`~escnn.group.Group.subgroup` to generate the subgroup.
        """
        pass
    
    @property
    @abstractmethod
    def subgroup_self_id(self):
        r"""
            The subgroup `id` associated with the group itself.
            The id can be used in the method :meth:`~escnn.group.Group.subgroup` to generate the subgroup.
        """
        pass

    @property
    @abstractmethod
    def _keys(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def generators(self) -> List[GroupElement]:
        r"""
            If the group is finite (``self.continuous = False``), a list of group elements which can generate this group.
            Should raise a `ValueError` if the group is not finite.
            
            Returns:
                a list of :class:`~escnn.group.GroupElement` instances
        """
        pass
    
    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    @abstractmethod
    def _combine(self, e1, e2,
                 param: str,
                 param1: str = None,
                 param2: str = None
                 ):
        r"""

        Method that returns the combination of two group elements according to the *group law*.
        
        Args:
            e1: an element of the group
            e2: another element of the group
    
        Returns:
            the group element :math:`e_1 \cdot e_2`
            
        """
        pass

    @abstractmethod
    def _inverse(self, element, param: str):
        r"""
        Method that returns the inverse in the group of the element given as input

        Args:
            element: an element of the group

        Returns:
            its inverse
        """
        pass

    @abstractmethod
    def _is_element(self,
                    element,
                    param: str,
                    verbose: bool = False
                    ) -> bool:
        r"""
        Check whether the input is an element of this group or not.

        Args:
            element: input object to test

        Returns:
            if the input is an element of the group

        """
        pass

    @abstractmethod
    def _equal(self, e1, e2,
               param: str,
               param1: str = None,
               param2: str = None
               ) -> bool:
        r"""
        Method that checks whether the two inputs are the same element of the group.

        This is especially useful for continuous groups with periodicity; see for instance
        :meth:`escnn.group.SO2.equal`.

        Args:
            e1: an element of the group
            e2: another element of the group

        Returns:
            if they are equal

        """
        pass

    @abstractmethod
    def _change_param(self, element, p_from: str, p_to: str):
        pass

    @abstractmethod
    def _hash_element(self, element, param: str):
        r"""
        Method that returns a unique hash for a group element given in input

        Args:
            element: an element of the group

        Returns:
            a unique hash
        """
        pass
    
    @abstractmethod
    def _repr_element(self, element, param: str):
        r"""
        Method that returns a representative string for a group element given in input

        Args:
            element: an element of the group

        Returns:
            a unique hash
        """
        pass

    ###########################################################################

    def __repr__(self):
        return self.name
    
    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def sample(self) -> GroupElement:
        r"""
            Sample a random element of the group from a uniform distribution over the group.
           
            Returns:
                :class:`~escnn.group.GroupElement`: the element sampled
            
        """
        pass

    def grid(self, *args, **kwargs) -> List[GroupElement]:
        r"""
        Method to generate collections fo points over the group.
        Each group should implement its own set of collections.
        Check the individual groups' documentations for details about the supported arguments.

        Returns:
            a list of :class:`~escnn.group.GroupElement` instances

        """
        raise NotImplementedError()

    def _process_subgroup_id(self, id):
        return id

    def subgroup(self, id) -> Tuple[
        escnn.group.Group,
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement],
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement]
    ]:
        r"""
        Restrict the current group to the subgroup identified by the input ``id``.

        Args:
            id: the identifier of the subgroup

        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)

        """
        id = self._process_subgroup_id(id)
        
        if id not in self._subgroups:
            subgroup, parent_mapping, child_mapping = self._subgroup(id)
            self._subgroups[id] = subgroup, parent_mapping, child_mapping

        return self._subgroups[id]

    @abstractmethod
    def _subgroup(self, id) -> Tuple[
        escnn.group.Group,
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement],
        Callable[[escnn.group.GroupElement], escnn.group.GroupElement]
    ]:
        r"""
        Restrict the current group to the subgroup identified by the input ``id``.

        Args:
            id: the identifier of the subgroup

        Returns:
            a tuple containing

                - the subgroup,

                - a function which maps an element of the subgroup to its inclusion in the original group and

                -a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)

        """
        pass

    def _combine_subgroups(self, sg_id1, sg_id2):
        raise NotImplementedError
    
    def irreps(self) -> List[escnn.group.IrreducibleRepresentation]:
        r"""
        List containing all irreducible representations (:class:`~escnn.group.IrreducibleRepresentation`)
        currently instantiated for this group.

        Returns:
            a list containing all irreducible representations built

        """
        return list(self._irreps.values())

    @property
    def representations(self) -> Dict[str, escnn.group.Representation]:
        r"""
        Dictionary containing all representations (:class:`~escnn.group.Representation`)
        instantiated for this group.

        Returns:
            a dictionary containing all representations built

        """
        return self._representations
    
    @property
    @abstractmethod
    def trivial_representation(self) -> escnn.group.IrreducibleRepresentation:
        r"""
        Builds the trivial representation of the group.
        The trivial representation is a 1-dimensional representation which maps any element to 1,
        i.e. :math:`\forall g \in G,\ \rho(g) = 1`.
        
        Returns:
            the trivial representation of the group

        """
        pass

    @abstractmethod
    def irrep(self, *id) -> escnn.group.IrreducibleRepresentation:
        r"""

        Builds the irreducible representation (:class:`~escnn.group.IrreducibleRepresentation`) of the group which is
        specified by the input arguments.

        .. seealso ::

            Check the documentation of the specific group subclass used for more information on the valid ``id`` values.

        Args:
            *id: parameters identifying the specific irrep.

        Returns:
            the irrep built

        """
        # TODO implement memoization here and let subclasses define an _irrep(*id) module
        pass

    @property
    def regular_representation(self) -> escnn.group.Representation:
        r"""
        Builds the regular representation of the group if the group has a *finite* number of elements;
        returns ``None`` otherwise.
        
        The regular representation of a finite group :math:`G` acts on a vector space :math:`\R^{|G|}` by permuting its
        axes.
        Specifically, associating each axis :math:`e_g` of :math:`\R^{|G|}` to an element :math:`g \in G`, the
        representation of an element :math:`\tilde{g}\in G` is a permutation matrix which maps :math:`e_g` to
        :math:`e_{\tilde{g}g}`.
        For instance, the regular representation of the group :math:`C_4` with elements
        :math:`\{r^k | k=0,\dots,3 \}` is instantiated by:
        
        +-----------------------------------+------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------+
        |    :math:`g`                      |          :math:`e`                                                                                         |          :math:`r`                                                                                         |        :math:`r^2`                                                                                         |        :math:`r^3`                                                                                         |
        +===================================+============================================================================================================+============================================================================================================+============================================================================================================+============================================================================================================+
        |  :math:`\rho_\text{reg}^{C_4}(g)` | :math:`\begin{bmatrix} 1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \\  0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 1 \\ \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 0 & 1 \\ 1 & 0 & 0 & 0 \\  0 & 1 & 0 & 0 \\ 0 & 0 & 1 & 0 \\ \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 0 & 1 & 0 \\ 0 & 0 & 0 & 1 \\  1 & 0 & 0 & 0 \\ 0 & 1 & 0 & 0 \\ \end{bmatrix}` | :math:`\begin{bmatrix} 0 & 1 & 0 & 0 \\ 0 & 0 & 1 & 0 \\  0 & 0 & 0 & 1 \\ 1 & 0 & 0 & 0 \\ \end{bmatrix}` |
        +-----------------------------------+------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------+------------------------------------------------------------------------------------------------------------+
        
        A vector :math:`v=\sum_g v_g e_g` in :math:`\R^{|G|}` can be interpreted as a scalar function
        :math:`v:G \to \R,\, g \mapsto v_g` on :math:`G`.
        
        Returns:
            the regular representation of the group

        """
        if self.order() < 0:
            raise ValueError(f"Regular representation is supported only for finite groups but "
                             f"the group {self.name} has an infinite number of elements")
        else:
            if "regular" not in self.representations:
                irreps, change_of_basis, change_of_basis_inv = escnn.group.representation.build_regular_representation(self)
                supported_nonlinearities = ['pointwise', 'norm', 'gated', 'concatenated']
                self.representations["regular"] = escnn.group.Representation(self,
                                                                             "regular",
                                                                             [r.id for r in irreps],
                                                                             change_of_basis,
                                                                             supported_nonlinearities,
                                                                             change_of_basis_inv=change_of_basis_inv,
                                                                             )
            return self.representations["regular"]

    def quotient_representation(self,
                                subgroup_id,
                                representatives: List[GroupElement] = None,
                                name: str = None,
                                ) -> escnn.group.Representation:
        r"""
        Builds the quotient representation of the group with respect to the subgroup identified by the
        input ``subgroup_id``.
        
        Similar to :meth:`~escnn.group.Group.regular_representation`, the quotient representation
        :math:`\rho_\text{quot}^{G/H}` of :math:`G` w.r.t. a subgroup :math:`H` acts on :math:`\R^{|G|/|H|}` by
        permuting its axes.
        Labeling the axes by the cosets :math:`gH` in the quotient space :math:`G/H`, it can be defined via its action
        :math:`\rho_\text{quot}^{G/H}(\tilde{g})e_{gH}=e_{\tilde{g}gH}`.

        Regular and trivial representations are two specific cases of quotient representations obtained by choosing
        :math:`H=\{e\}` or :math:`H=G`, respectively.
        Vectors in the representation space :math:`\R^{|G|/|H|}` can be viewed as scalar functions on the quotient
        space :math:`G/H`.
        
        The quotient representation :math:`\rho_\text{quot}^{G/H}` can also be defined as the
        :meth:`~escnn.group.Group.induced_representation` from the trivial representation of the subgroup :math:`H`.
        
        .. todo ::
            docs for `representatives`
        
        
        Args:
            subgroup_id: identifier of the subgroup
            representatives (list, optional):
            name (str, optional): optionally, specify a custom name for this representation
        
        Returns:
            the quotient representation of the group

        """

        if name is None:
            name = f"quotient[{subgroup_id}]"
        
        if name not in self.representations:
            subgroup, _, _ = self.subgroup(subgroup_id)
            
            supported_nonlinearities = _induced_nonlinearities(subgroup.trivial_representation)
            
            irreps, change_of_basis, change_of_basis_inv = self._induced_from_irrep(
                subgroup_id,
                subgroup.trivial_representation,
                representatives
            )
            self.representations[name] = escnn.group.Representation(self,
                                                                    name,
                                                                    [r.id for r in irreps],
                                                                    change_of_basis,
                                                                    supported_nonlinearities,
                                                                    change_of_basis_inv=change_of_basis_inv,
                                                                    )

        return self.representations[name]

    def induced_representation(self,
                               subgroup_id,
                               repr: escnn.group.IrreducibleRepresentation,
                               representatives: List[GroupElement] = None,
                               name: str = None
                               ) -> escnn.group.Representation:
        r"""
        Builds the induced representation from the input representation ``repr`` of the subgroup identified by
        the input ``subgroup_id``.
        
        .. todo ::
            docs for `representatives`
        
        Args:
            subgroup_id: identifier of the subgroup
            repr (Representation): the representation of the subgroup
            representatives (list, optional):
            name (str, optional): optionally, specify a custom name for this representation

        Returns:
            the induced representation of the group

        """
        
        assert repr.irreducible, "Induction from general representations is not supported yet"

        if name is None:
            name = f"induced[{subgroup_id}][{repr.name}]"

        if name not in self.representations:

            supported_nonlinearities = _induced_nonlinearities(repr)

            irreps, change_of_basis, change_of_basis_inv = self._induced_from_irrep(
                subgroup_id,
                repr,
                representatives
            )
            self.representations[name] = escnn.group.Representation(self,
                                                                    name,
                                                                    [r.id for r in irreps],
                                                                    change_of_basis,
                                                                    supported_nonlinearities,
                                                                    change_of_basis_inv=change_of_basis_inv,
                                                                    )

        return self.representations[name]

    def _induced_from_irrep(self, subgroup_id: Tuple[float, int],
                            repr: escnn.group.IrreducibleRepresentation,
                            representatives: List[GroupElement] = None,
                            ) -> Tuple[List[escnn.group.IrreducibleRepresentation], np.ndarray, np.ndarray]:
    
        r"""
        Builds the induced representation from the input *irreducible* representation ``repr`` of the subgroup
        identified by the input ``subgroup_id``.
        
        .. todo ::
            docs for `representatives`
        
        
        Args:
            subgroup_id: identifier of the subgroup
            repr (Representation): the representation of the subgroup
            

        Returns:
            a tuple containing the list of irreps, the change of basis and the inverse change of basis of
            the induced representation

        """
    
        assert repr.irreducible
        return escnn.group.representation.build_induced_representation(
            self,
            subgroup_id,
            repr,
            representatives
        )

    def spectral_regular_representation(self, *irreps, name: str = None) -> 'Representation':
        r"""
        Finite dimensional invariant subspace of the regular representation containing only the irreps passed in input.
        The regular representation is expressed in the spectral basis, i.e. as a direct sum of irreps.

        The optional parameter ``name`` is also used for caching purpose.
        Consecutive calls of this method using the same ``name`` will ignore the argument  ``irreps``
        and return the same instance of representation.

        .. seealso::
            :meth:`escnn.group.HomSpace.induced_representation`

        """

        if name is None:
            irreps_names = '|'.join(str(i) for i in irreps)
            name = f'regular_[{irreps_names}]'

        return self.spectral_quotient_representation(self.subgroup_trivial_id, *irreps, name=name)

    def spectral_quotient_representation(self, subgroup_id: Tuple, *irreps, name: str = None) -> 'Representation':
        r"""
        Finite dimensional invariant subspace of the quotient representation containing only the irreps passed in input.
        The quotient representation is expressed in the spectral basis, i.e. as a direct sum of irreps.

        The optional parameter ``name`` is also used for caching purpose.
        Consecutive calls of this method using the same ``name`` will ignore the arguments ``subgroup_id`` and ``irreps``
        and return the same instance of representation.

        .. seealso::
            :meth:`escnn.group.HomSpace.induced_representation`

        """

        if name is None:
            irreps_names = '|'.join(str(i) for i in irreps)
            name = f'quotient[{subgroup_id}]_[{irreps_names}]'

        if name not in self._representations:
            homspace = self.homspace(subgroup_id)
            self._representations[name] = homspace.induced_representation(homspace.H.trivial_representation.id, irreps, name)

        return self._representations[name]

    def restrict_representation(self, id, repr: escnn.group.Representation) -> escnn.group.Representation:
        r"""

        Restrict the input :class:`~escnn.group.Representation` to the subgroup identified by ``id``.
        
        Any representation :math:`\rho : G \to \GL{\R^n}` can be uniquely restricted to a representation
        of a subgroup :math:`H < G` by restricting its domain of definition:

        .. math ::

            \Res{H}{G}(\rho): H \to \GL{{\R}^n},\ h \mapsto \rho\big|_H(h)
        
        We recommend directly using the method :meth:`escnn.group.Representation.restrict`.
        
        .. seealso ::

            Check the documentation of the method :meth:`~escnn.group.Group.subgroup()` of the group used to see
            the available subgroups and accepted ids.

        Args:
            id: identifier of the subgroup
            repr (Representation): the representation to restrict

        Returns:
            the restricted representation

        """
    
        assert repr.group == self
    
        sg, _, _ = self.subgroup(id)
        id = self._process_subgroup_id(id)

        # First, restrict each irrep in the representation
    
        irreps_changes_of_basis = []
        irreps = []
    
        for irr in repr.irreps:
            irrep_cob, reduced_irreps = self._restrict_irrep(irr, id)
            size = self.irrep(*irr).size
            assert irrep_cob.shape == (size, size)
        
            irreps_changes_of_basis.append(irrep_cob)
            irreps += reduced_irreps
    
        # concatenate the restricted irreps and merge the representation's change of basis with the
        # restricted irreps' change of basis matrices
        irreps_changes_of_basis = sparse.block_diag(irreps_changes_of_basis, format='csc')
        change_of_basis = repr.change_of_basis @ irreps_changes_of_basis
    
        name = f"{self.name}:{repr.name}"
    
        resr = escnn.group.Representation(sg,
                                          name,
                                          irreps,
                                          change_of_basis,
                                          repr.supported_nonlinearities)
    
        if resr.is_trivial() and 'pointwise' not in repr.supported_nonlinearities:
            resr.supported_nonlinearities.add("pointwise")
    
        return resr

    def homspace(self, id) -> escnn.group.HomSpace:
        r"""
            If :math:`G` is the current group and ``id`` identifies the subgroup :math:`H`
            (see :meth:`~escnn.group.Group.subgroup`), this method generates the homogeneous space
            :class:`~escnn.group.HomSpace` :math:`X = G / H`.
            
            .. note ::

                The generated instance of :class:`~escnn.group.HomSpace` is cached inside the instance of the current
                group such that repeated calls of this method using the same ``id`` return the same instance of
                :class:`~escnn.group.HomSpace` and no additional computations are required.
            
            
        Returns:
            an instance of :class:`~escnn.group.HomSpace`

        """
        id = self._process_subgroup_id(id)
        
        if id not in self._homspaces:
            self._homspaces[id] = escnn.group.HomSpace(self, self._process_subgroup_id(id))
            
        return self._homspaces[id]
    
    @abstractmethod
    def _restrict_irrep(self, irrep: Tuple, id) -> Tuple[np.matrix, List[Tuple]]:
        pass

    def _clebsh_gordan_coeff(self, m, n, j) -> np.ndarray:
        group_keys = self._keys
        m = self.get_irrep_id(m)
        n = self.get_irrep_id(n)
        j = self.get_irrep_id(j)
        return escnn.group._clebsh_gordan._clebsh_gordan_tensor(m, n, j, self.__class__.__name__, **group_keys)

    def _tensor_product_irreps(self, m, n) -> List[Tuple[Tuple, int]]:
        group_keys = self._keys
        m = self.get_irrep_id(m)
        n = self.get_irrep_id(n)
        return escnn.group._clebsh_gordan._find_tensor_decomposition(m, n, self.__class__.__name__, **group_keys)

    def _tensor_product(self, rho1: escnn.group.Representation, rho2: escnn.group.Representation) -> escnn.group.Representation:
        assert rho1.group == self
        assert rho2.group == self
        
        D1 = rho1.size
        D2 = rho2.size
        D = D1 * D2
        
        change_of_basis = np.zeros((D, D))
        irreps = []
        
        p = 0
        for irr1 in rho1.irreps:
            irr1 = self.irrep(*irr1)
            
            permutation = np.zeros((irr1.size * rho2.size, irr1.size * rho2.size))
            q = 0
            
            for irr2 in rho2.irreps:
                irr2 = self.irrep(*irr2)

                irr1_tensor_irr2 = self._tensor_product_irreps(irr1.id, irr2.id)
                size = 0
                for irr_id, S in irr1_tensor_irr2:
                    irr = self.irrep(*irr_id)
                    size += irr.size*S
                    irreps += [irr.id]*S
                
                assert size == irr1.size * irr2.size, (size, irr1.size, irr2.size)

                i = 0
                for irr_j, S in irr1_tensor_irr2:
                    irr_j = self.irrep(*irr_j)
                    change_of_basis[
                        p:p+size,
                        p+i:p+i+irr_j.size*S
                    ] = self._clebsh_gordan_coeff(irr1.id, irr2.id, irr_j.id).reshape(-1, irr_j.size*S)
                    i += irr_j.size * S
                
                assert i == size
                
                for i in range(irr1.size):
                    permutation[
                        q + i*rho2.size:q + i*rho2.size + irr2.size,
                        q*irr1.size + i * irr2.size:q*irr1.size + (i+1) * irr2.size
                    ] = np.eye(irr2.size)
                
                q += irr2.size
                p += size
            
            assert np.allclose(permutation @ permutation.T, np.eye(permutation.shape[0]))
            assert np.allclose(permutation.T @ permutation, np.eye(permutation.shape[0]))

            change_of_basis[
                p - irr1.size*rho2.size:p,
                p - irr1.size*rho2.size:p
            ] = permutation @ change_of_basis[p - irr1.size*rho2.size:p, p - irr1.size*rho2.size:p]
            
        change_of_basis = np.kron(rho1.change_of_basis, rho2.change_of_basis) @ change_of_basis

        assert p == sum(self.irrep(*irr).size for irr in irreps), (p, rho1.size, rho2.size)
        assert p == rho1.size * rho2.size, (p, rho1.size, rho2.size)
        assert p == change_of_basis.shape[0]
        assert p == change_of_basis.shape[1]

        supported_nonlinearities = _tensor_nonlinearities(rho1, rho2)

        character = _tensor_product_character(rho1, rho2)
        
        if len(irreps) > 1:
            return escnn.group.Representation(self,
                                              f'{rho1.name} X {rho2.name}',
                                              irreps,
                                              change_of_basis,
                                              character=character,
                                              supported_nonlinearities=supported_nonlinearities
                                              )
        else:
            return escnn.group.change_basis(
                self.irrep(*irreps[0]),
                change_of_basis,
                name=f'{rho1.name} X {rho2.name}'
            )

    @abstractmethod
    def testing_elements(self) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        """
        pass

    def get_irrep_id(self, psi):
        if isinstance(psi, escnn.group.IrreducibleRepresentation):
            assert psi.group == self
            return psi.id
        elif isinstance(psi, str):
            psi = self.representations[psi]
            assert isinstance(psi, escnn.group.IrreducibleRepresentation)
            return psi
        elif isinstance(psi, tuple):
            return self.irrep(*psi).id
        else:
            return self.irrep(psi).id

    def _decode_subgroup_id_pickleable(self, id: Tuple) -> Tuple:

        if isinstance(id, tuple):
            if id[0] == 'GROUPELEMENT':
                id = self.element(id[1], id[2])
            else:
                id = list(id)
                for i in range(len(id)):
                    id[i] = self._decode_subgroup_id_pickleable(id[i])
                id = tuple(id)

        return id

    def _encode_subgroup_id_pickleable(self, id: Tuple) -> Tuple:
        if isinstance(id, GroupElement):
            id = 'GROUPELEMENT', id.value, id.param
        elif isinstance(id, tuple):
            id = list(id)
            for i in range(len(id)):
                id[i] = self._encode_subgroup_id_pickleable(id[i])
            id = tuple(id)
        return id

    @classmethod
    @abstractmethod
    def _generator(cls, *args, **kwargs) -> 'Group':
        # TODO solve the singleton problem!!!
        pass


def _tensor_product_character(rho1: 'Representation', rho2: 'Representation'):
    
    def character(e: GroupElement, rho1=rho1, rho2=rho2) -> float:
        return rho1.character(e) * rho2.character(e)
    
    return character


def _induced_nonlinearities(repr: escnn.group.Representation):
    
    supported_nonlinearities = []
    
    if 'pointwise' in repr.supported_nonlinearities:
        supported_nonlinearities.append('pointwise')
    if 'concatenated' in repr.supported_nonlinearities:
        supported_nonlinearities.append('concatenated')
    if 'gated' in repr.supported_nonlinearities:
        supported_nonlinearities.append('gated')
        for nl in repr.supported_nonlinearities:
            if nl.startswith('induced_gated'):
                supported_nonlinearities.append(nl)
                break
        else:
            supported_nonlinearities.append(f'induced_gated_{repr.size}')
    if 'norm' in repr.supported_nonlinearities:
        supported_nonlinearities.append('norm')
        for nl in repr.supported_nonlinearities:
            if nl.startswith('induced_norm'):
                supported_nonlinearities.append(nl)
                break
        else:
            supported_nonlinearities.append(f'induced_norm_{repr.size}')
    if 'gate' in repr.supported_nonlinearities or 'induced_gate' in repr.supported_nonlinearities:
        supported_nonlinearities.append('induced_gate')

    # 'vectorfield' not always supported by the induced representation so they are not added

    return supported_nonlinearities


def _tensor_nonlinearities(repr1: escnn.group.Representation, repr2: escnn.group.Representation):
    supported_nonlinearities = []
    
    if 'pointwise' in repr1.supported_nonlinearities and 'pointwise' in repr2.supported_nonlinearities:
        supported_nonlinearities.append('pointwise')
        
    supported_nonlinearities.append('gated')
    supported_nonlinearities.append('norm')

    if 'gate' in repr1.supported_nonlinearities and 'gate' in repr2.supported_nonlinearities:
        supported_nonlinearities.append('gate')
        
    # TODO - check for induced non-linearities
    
    return supported_nonlinearities


class GroupElement(ABC):
    
    def __init__(self, g, group: Group, param: str = None):
        r"""
        
        Class implementing an element of a group.
        
        Group elements can be combined the group operations like the *group law* or the *inverse*.
        In particular, one can combine two group elements through the group law using the
        operator ``@`` or compute the inverse of an element using ``~``.
        For example ::
            
            G = so3_group()
            a = G.sample()
            b = G.sample()
            
            c = a @ b
            a_ = ~a
            
            e = G.identity
            
            assert e == ~e
            assert a == a @ e
            
            
        Args:
            g: values parametrizing the group element
            group (Group): the group this element belongs to
            param (str): the type of parametrization of ``g``
            
        Attributes:
            ~.group (Group): the group it belongs to
            
        """

        if param is None:
            param = group.PARAM
        
        # TODO - create "ParametrizationError"?
        if param not in group.PARAMETRIZATIONS:
            raise ValueError(f'Error! {param} not recognized. Expected one of {group.PARAMETRIZATIONS}')
        
        if not group._is_element(g, param): #, verbose=True):
            raise ValueError(f'Error! {g} is not an element of {group} under the parametrization [{param}]')
        
        # Group: the group this element belongs to
        self.group = group
        
        # TODO: do lazy conversion. Keep always input parametrization.
        #  Convert to self.group.PARAM only when performing operations and creating new group elements
        self._element = group._change_param(g, param, group.PARAM)
    
    def __eq__(self, other: GroupElement):
        if not isinstance(other, GroupElement) or other.group != self.group:
            return False
        return self.group._equal(self._element, other._element, param1=self.param, param2=other.param)

    def __matmul__(self, other: GroupElement):
        if isinstance(other, GroupElement):
            if other.group == self.group:
                return GroupElement(
                    self.group._combine(self._element, other._element, param1=self.param, param2=other.param),
                    self.group,
                    self.param
                )
            else:
                raise NotImplementedError(f'Multiplication of group elements which belong to different groups is not supported.')
        else:
            return NotImplemented

    def __invert__(self):
        return GroupElement(
            self.group._inverse(self._element, param=self.param),
            self.group,
            self.param
        )

    def __hash__(self):
        return self.group._hash_element(self._element, self.param)

    def __repr__(self):
        return self.group._repr_element(self._element, self.param)

    @property
    def value(self):
        r"""
            Returns the values of the internal parametrization of the group element.
            These values parametrize the group element according to the parametrization
            :attr:`escnn.group.GroupElement.param`.

        """
        return self._element

    @property
    def param(self) -> str:
        r"""
            The type parametrization used internally to store this group element.
        """
        return self.group.PARAM

    def to(self, param: str):
        r"""
            Converts the current group element to the input parametrization ``param`` and returns the corresponding
            values.

            .. note ::
                This method does *not* return an instance of :class:`~escnn.group.GroupElement`.
                This method does *not* affect the internal representation of the element, but just converts it to the
                input ``param`` and returns the converted values.

        """
        return self.group._change_param(self._element, self.param, param)

